"""
This is an experimental implementation of a restricted input field.
If the implementation is successful, the feature will be merged into the existing UIInputText class.
"""

from arcade.gui import UIEvent, UIInputText


class UIRestrictedInput(UIInputText):
    """
    A text input field that restricts the input to a certain type.

    This class is meant to be subclassed to create custom input fields
     that restrict the input by providing a custom validation method.

     Invalid inputs are dropped.
    """

    @property
    def text(self):
        """Text of the input field."""
        return self.doc.text

    @text.setter
    def text(self, text: str):
        if not self.validate(text):
            # if the text is invalid, do not update the text
            return

        # we can not call super().text = text here: https://bugs.python.org/issue14965
        UIInputText.text.__set__(self, text)  # type: ignore

    def on_event(self, event: UIEvent) -> bool | None:
        # check if text changed during event handling,
        # if so we need to validate the new text
        old_text = self.text
        pos = self.caret.position

        result = super().on_event(event)
        if not self.validate(self.text):
            self.text = old_text
            self.caret.position = pos

        return result

    def validate(self, text) -> bool:
        """Override this method to add custom validation logic.

        Be aware that an empty string should always be valid.
        """
        return True


class UIIntInput(UIRestrictedInput):
    def validate(self, text) -> bool:
        if text == "":
            return True

        try:
            int(text)
            return True
        except ValueError:
            return False


class UIFloatInput(UIRestrictedInput):
    def validate(self, text) -> bool:
        if text == "":
            return True

        try:
            float(text)
            return True
        except ValueError:
            return False


class UIRegexInput(UIRestrictedInput):
    def __init__(self, *args, pattern: str = r".*", **kwargs):
        super().__init__()
        self.pattern = pattern

    def validate(self, text: str) -> bool:
        if text == "":
            return True

        import re

        return re.match(self.pattern, text) is not None



# type: ignore
from __future__ import annotations

from collections.abc import Callable
from enum import Enum
from typing import Any, TypeVar

import pyglet
from pyglet.input.base import Controller
from typing_extensions import TypedDict

import arcade
from arcade.future.input import inputs
from arcade.future.input.input_mapping import (
    Action,
    ActionMapping,
    Axis,
    AxisMapping,
    serialize_action,
    serialize_axis,
)
from arcade.future.input.inputs import InputEnum, InputType
from arcade.future.input.raw_dicts import RawAction, RawAxis
from arcade.types import OneOrIterableOf
from arcade.utils import grow_sequence


class RawInputManager(TypedDict):
    actions: list[RawAction]
    axes: list[RawAxis]
    controller_deadzone: float


_K = TypeVar("_K")
_V = TypeVar("_V")


def _clean_dicts(to_remove: _V, *dicts_to_clean: dict[_K, set[_V]]) -> None:
    """Clean a dictionary in-place.

    This helps simplify serialization code for controller config.
    """
    to_discard = []
    for to_clean in dicts_to_clean:
        if to_discard:
            to_discard.clear()

        for key, set_value in to_clean.items():
            set_value.discard(to_remove)
            if len(set_value) == 0:
                to_discard.append(key)

        for key in to_discard:
            del to_clean[key]


class ActionState(Enum):
    PRESSED = 1
    RELEASED = 0


class InputDevice(Enum):
    KEYBOARD = 0
    CONTROLLER = 1


class InputManager:
    def __init__(
        self,
        controller: Controller | None = None,
        allow_keyboard: bool = True,
        action_handlers: OneOrIterableOf[Callable[[str, ActionState], Any]] = tuple(),
        controller_deadzone: float = 0.1,
    ):
        self.actions: dict[str, Action] = {}

        # We don't use defaultdict here since:
        # 1. These attributes are unprotected
        # 2. That means sets would be created on *any* access by the user
        # 2. Those leftover sets would complicate serialization
        self.keys_to_actions: dict[int, set[str]] = {}
        self.controller_buttons_to_actions: dict[str, set[str]] = {}
        self.controller_axes_to_actions: dict[str, set[str]] = {}
        self.mouse_buttons_to_actions: dict[int, set[str]] = {}

        self.on_action_listeners: list[Callable[[str, ActionState], Any]] = []
        self.action_subscribers: dict[str, set[Callable[[ActionState], Any]]] = {}

        self.axes: dict[str, Axis] = {}
        self.axes_state: dict[str, float] = {}
        self.keys_to_axes: dict[int, set[str]] = {}
        self.controller_buttons_to_axes: dict[str, set[str]] = {}
        self.controller_analog_to_axes: dict[str, set[str]] = {}

        self.window = arcade.get_window()

        self.register_action_handler(action_handlers)

        self._allow_keyboard = allow_keyboard
        if self._allow_keyboard:
            self.window.push_handlers(
                self.on_key_press,
                self.on_key_release,
                self.on_mouse_press,
                self.on_mouse_release,
            )

        self.active_device = None

        if self._allow_keyboard:
            self.active_device = InputDevice.KEYBOARD

        self.controller = None
        self.controller_deadzone = controller_deadzone
        if controller:
            self.controller = controller
            self.controller.open()
            self.controller.push_handlers(
                self.on_button_press,
                self.on_button_release,
                self.on_stick_motion,
                self.on_dpad_motion,
                self.on_trigger_motion,
            )
            self.active_device = InputDevice.CONTROLLER

    def serialize(self) -> RawInputManager:
        raw_actions = []
        for action in self.actions.values():
            raw_actions.append(serialize_action(action))
        raw_axes = []
        for axis in self.axes.values():
            raw_axes.append(serialize_axis(axis))
        return {
            "actions": raw_actions,
            "axes": raw_axes,
            "controller_deadzone": self.controller_deadzone,
        }

    @classmethod
    def parse(cls, raw: RawInputManager) -> InputManager:
        final = cls(controller_deadzone=raw["controller_deadzone"])

        for raw_action in raw["actions"]:
            name = raw_action["name"]
            final.new_action(name)

            for raw_mapping in raw_action["mappings"]:
                input_instance = inputs.parse_instance(raw_mapping)

                final.add_action_input(
                    name,
                    input_instance,
                    raw_mapping["mod_shift"],
                    raw_mapping["mod_ctrl"],
                    raw_mapping["mod_alt"],
                )

        for raw_axis in raw["axes"]:
            name = raw_axis["name"]
            final.new_axis(name)
            for raw_mapping in raw_axis["mappings"]:
                input_instance = inputs.parse_instance(raw_mapping)

                final.add_axis_input(name, input_instance, raw_mapping["scale"])

        return final

    def copy_existing(self, existing: InputManager):
        self.actions = existing.actions.copy()
        self.keys_to_actions = existing.keys_to_actions.copy()
        self.controller_buttons_to_actions = existing.controller_buttons_to_actions.copy()
        self.mouse_buttons_to_actions = existing.mouse_buttons_to_actions.copy()
        self.axes = existing.axes.copy()
        self.axes_state = existing.axes_state.copy()
        self.controller_buttons_to_axes = existing.controller_buttons_to_axes.copy()
        self.controller_analog_to_axes = existing.controller_analog_to_axes.copy()
        self.controller_deadzone = existing.controller_deadzone

    @classmethod
    def from_existing(
        cls,
        existing: InputManager,
        controller: pyglet.input.Controller | None = None,
    ) -> InputManager:
        new = cls(
            allow_keyboard=existing.allow_keyboard,
            controller=controller,
            controller_deadzone=existing.controller_deadzone,
        )
        new.copy_existing(existing)
        new.actions = existing.actions.copy()

        return new

    def bind_controller(self, controller: Controller):
        if self.controller:
            self.controller.remove_handlers()

        self.controller = controller
        self.controller.open()
        self.controller.push_handlers(
            self.on_button_press,
            self.on_button_release,
            self.on_stick_motion,
            self.on_dpad_motion,
            self.on_trigger_motion,
        )
        self.active_device = InputDevice.CONTROLLER

    def unbind_controller(self):
        if not self.controller:
            return

        self.controller.remove_handlers(
            self.on_button_press,
            self.on_button_release,
            self.on_stick_motion,
            self.on_dpad_motion,
            self.on_trigger_motion,
        )
        self.controller.close()
        self.controller = None

        if self._allow_keyboard:
            self.active_device = InputDevice.KEYBOARD

    @property
    def allow_keyboard(self):
        return self._allow_keyboard

    @allow_keyboard.setter
    def allow_keyboard(self, value: bool):
        if self._allow_keyboard == value:
            return

        self._allow_keyboard = value
        if self._allow_keyboard:
            self.window.push_handlers(
                self.on_key_press,
                self.on_key_release,
                self.on_mouse_press,
                self.on_mouse_release,
            )
        else:
            self.window.remove_handlers(self)

    def new_action(
        self,
        name: str,
    ):
        action = Action(name)
        self.actions[name] = action

    def remove_action(self, name: str):
        self.clear_action_input(name)

        to_remove = self.actions.get(name, None)
        if to_remove:
            del self.actions[name]

    def add_action_input(
        self,
        action: str,
        input: InputEnum,
        mod_shift: bool = False,
        mod_ctrl: bool = False,
        mod_alt: bool = False,
    ):
        mapping = ActionMapping(input, mod_shift, mod_ctrl, mod_alt)
        self.actions[action].add_mapping(mapping)

        if mapping._input_type == InputType.KEYBOARD:
            # input is guaranteed to be an instance of Keys enum at this point
            if input.value not in self.keys_to_actions:
                self.keys_to_actions[input.value] = set()
            self.keys_to_actions[input.value].add(action)
        elif mapping._input_type == InputType.CONTROLLER_BUTTON:
            if input.value not in self.controller_buttons_to_actions:
                self.controller_buttons_to_actions[input.value] = set()
            self.controller_buttons_to_actions[input.value].add(action)
        elif mapping._input_type == InputType.MOUSE_BUTTON:
            if input.value not in self.mouse_buttons_to_actions:
                self.mouse_buttons_to_actions[input.value] = set()
            self.mouse_buttons_to_actions[input.value].add(action)
        elif mapping._input_type == InputType.CONTROLLER_AXIS:
            if input.value not in self.controller_axes_to_actions:
                self.controller_axes_to_actions[input.value] = set()
            self.controller_axes_to_actions[input.value].add(action)

    def clear_action_input(self, action: str):
        self.actions[action]._mappings.clear()
        _clean_dicts(
            action,
            self.keys_to_actions,
            self.controller_buttons_to_actions,
            self.controller_axes_to_actions,
            self.mouse_buttons_to_actions,
        )

    def register_action_handler(self, handler: OneOrIterableOf[Callable[[str, ActionState], Any]]):
        grow_sequence(self.on_action_listeners, handler, append_if=callable)

    def subscribe_to_action(self, name: str, subscriber: Callable[[ActionState], Any]):
        old = self.action_subscribers.get(name, set())
        old.add(subscriber)
        self.action_subscribers[name] = old

    def new_axis(self, name: str):
        if name in self.axes:
            raise AttributeError(f"Tried to create Axis with duplicate name: {name}")

        axis = Axis(name)
        self.axes[name] = axis
        self.axes_state[name] = 0.0

    def add_axis_input(self, axis: str, input: InputEnum, scale: float = 1.0):
        mapping = AxisMapping(input, scale)
        self.axes[axis].add_mapping(mapping)

        if mapping._input_type == InputType.KEYBOARD:
            if input.value not in self.keys_to_axes:
                self.keys_to_axes[input.value] = set()
            self.keys_to_axes[input.value].add(axis)
        elif mapping._input_type == InputType.CONTROLLER_BUTTON:
            if input.value not in self.controller_buttons_to_axes:
                self.controller_buttons_to_axes[input.value] = set()
            self.controller_buttons_to_axes[input.value].add(axis)
        elif mapping._input_type == InputType.CONTROLLER_AXIS:
            if input.value not in self.controller_analog_to_axes:
                self.controller_analog_to_axes[input.value] = set()
            self.controller_analog_to_axes[input.value].add(axis)

    def clear_axis_input(self, axis: str):
        self.axes[axis]._mappings.clear()
        _clean_dicts(
            axis, self.keys_to_axes, self.controller_analog_to_axes, self.controller_buttons_to_axes
        )

    def remove_axis(self, name: str):
        self.clear_axis_input(name)

        to_remove = self.axes.get(name, None)
        if to_remove:
            del self.axes[name]
            del self.axes_state[name]

    def axis(self, name: str) -> float:
        return self.axes_state[name]

    def dispatch_action(self, action: str, state: ActionState):
        arcade.get_window().dispatch_event("on_action", action, state)
        for listener in self.on_action_listeners:
            listener(action, state)
        if action in self.action_subscribers:
            for subscriber in tuple(self.action_subscribers[action]):
                subscriber(state)

    def on_mouse_press(self, x: int, y: int, button: int, modifiers: int) -> None:
        if not self._allow_keyboard:
            return

        self.active_device = InputDevice.KEYBOARD
        mouse_buttons_to_actions = tuple(self.mouse_buttons_to_actions.get(button, set()))
        for action_name in mouse_buttons_to_actions:
            action = self.actions[action_name]
            hit = True
            for mapping in tuple(action._mappings):
                if mapping._modifiers:
                    for mod in mapping._modifiers:
                        if not modifiers & mod:
                            hit = False

            if hit:
                self.dispatch_action(action_name, ActionState.PRESSED)

    def on_key_press(self, key: int, modifiers: int) -> None:
        if not self._allow_keyboard:
            return

        self.active_device = InputDevice.KEYBOARD
        keys_to_actions = tuple(self.keys_to_actions.get(key, set()))
        for action_name in keys_to_actions:
            action = self.actions[action_name]
            hit = True
            for mapping in tuple(action._mappings):
                if mapping._modifiers:
                    for mod in mapping._modifiers:
                        if not modifiers & mod:
                            hit = False

            if hit:
                self.dispatch_action(action_name, ActionState.PRESSED)

    def on_mouse_release(self, x: int, y: int, button: int, modifiers: int) -> None:
        if not self._allow_keyboard:
            return

        mouse_buttons_to_actions = tuple(self.mouse_buttons_to_actions.get(button, set()))
        for action_name in mouse_buttons_to_actions:
            action = self.actions[action_name]
            hit = True
            for mapping in tuple(action._mappings):
                if mapping._modifiers:
                    for mod in mapping._modifiers:
                        if not modifiers & mod:
                            hit = False

            if hit:
                self.dispatch_action(action_name, ActionState.RELEASED)

    def on_key_release(self, key: int, modifiers: int) -> None:
        if not self._allow_keyboard:
            return
        # What, why are we doing any of this repeat tuple conversion in here?
        keys_to_actions = tuple(self.keys_to_actions.get(key, set()))
        for action_name in keys_to_actions:
            action = self.actions[action_name]
            hit = True
            for mapping in tuple(action._mappings):
                if mapping._modifiers:
                    for mod in mapping._modifiers:
                        if not modifiers & mod:
                            hit = False

            if hit:
                self.dispatch_action(action_name, ActionState.RELEASED)

    def on_button_press(self, controller: Controller, button_name: str):
        self.active_device = InputDevice.CONTROLLER
        buttons_to_actions = tuple(self.controller_buttons_to_actions.get(button_name, set()))
        for action_name in buttons_to_actions:
            self.dispatch_action(action_name, ActionState.PRESSED)

    def on_button_release(self, controller: Controller, button_name: str):
        buttons_to_actions = tuple(self.controller_buttons_to_actions.get(button_name, set()))
        for action_name in buttons_to_actions:
            self.dispatch_action(action_name, ActionState.RELEASED)

    def on_stick_motion(self, controller: Controller, name: str, motion: pyglet.math.Vec2):
        x_value, y_value = motion.x, motion.y
        if name == "leftx":
            self.window.dispatch_event(
                "on_stick_motion",
                self.controller,
                "leftxpositive" if x_value > 0 else "leftxnegative",
                x_value,
                y_value,
            )
        elif name == "lefty":
            self.window.dispatch_event(
                "on_stick_motion",
                self.controller,
                "leftypositive" if y_value > 0 else "leftynegative",
                x_value,
                y_value,
            )
        elif name == "rightx":
            self.window.dispatch_event(
                "on_stick_motion",
                self.controller,
                "rightxpositive" if x_value > 0 else "rightxpositive",
                x_value,
                y_value,
            )
        elif name == "righty":
            self.window.dispatch_event(
                "on_stick_motion",
                self.controller,
                "rightypositive" if y_value > 0 else "rightynegative",
                x_value,
                y_value,
            )

        axes_to_actions = self.controller_axes_to_actions.get(name, set())

        if (
            x_value > self.controller_deadzone
            or x_value < -self.controller_deadzone
            or y_value > self.controller_deadzone
            or y_value < -self.controller_deadzone
        ):
            self.active_device = InputDevice.CONTROLLER

            for action_name in axes_to_actions:
                self.dispatch_action(action_name, ActionState.PRESSED)

            return

        for action_name in axes_to_actions:
            self.dispatch_action(action_name, ActionState.RELEASED)

    def on_dpad_motion(self, controller: Controller, motion: pyglet.math.Vec2):
        self.active_device = InputDevice.CONTROLLER

    def on_trigger_motion(self, controller: Controller, trigger_name: str, value: float):
        self.active_device = InputDevice.CONTROLLER

    def update(self):
        for name in self.axes.keys():
            self.axes_state[name] = 0

        if self.controller and self.active_device == InputDevice.CONTROLLER:
            for name, axis in self.axes.items():
                for mapping in tuple(axis._mappings):
                    if mapping._input_type == InputType.CONTROLLER_AXIS:
                        scale = mapping._scale
                        input = getattr(self.controller, mapping._input.value)  # type: ignore
                        if input > self.controller_deadzone or input < -self.controller_deadzone:
                            self.axes_state[name] = input * scale
                    if mapping._input_type == InputType.CONTROLLER_BUTTON:
                        if getattr(self.controller, mapping._input.value):  # type: ignore
                            self.axes_state[name] = mapping._scale
        elif self.active_device == InputDevice.KEYBOARD and self._allow_keyboard:
            for name, axis in self.axes.items():
                for mapping in tuple(axis._mappings):
                    if mapping._input_type == InputType.KEYBOARD:
                        if self.window.keyboard[mapping._input.value]:
                            self.axes_state[name] = mapping._scale
                    elif mapping._input_type == InputType.MOUSE_AXIS:
                        self.axes_state[name] = (
                            self.window.mouse[mapping._input.name.lower()] * mapping._scale
                        )
                    elif mapping._input_type == InputType.MOUSE_BUTTON:
                        if self.window.mouse[mapping._input.value]:
                            self.axes_state[name] = mapping._scale



"""
A stationary Background.

This program loads a background from a file.
The texture is not added to a texture atlas or texture cache.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.background_stationary
"""

import arcade
import arcade.future.background as background

WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720

WINDOW_TITLE = "Stationary Background Example"

PLAYER_SPEED = 300
CAMERA_SPEED = 0.1


class GameView(arcade.View):
    def __init__(self):
        super().__init__()
        self.camera = arcade.Camera2D()

        # Load the background from file. It defaults to the size of the texture
        # with the bottom left corner at (0, 0).
        self.background = background.Background.from_file(
            ":resources:/images/backgrounds/abstract_1.jpg"
        )
        self.background.size = (WINDOW_WIDTH, WINDOW_HEIGHT)

        # Create the player sprite.
        self.player_sprite = arcade.SpriteSolidColor(20, 30, color=arcade.color.PURPLE)
        self.player_sprite.center_y = self.camera.viewport_height // 2
        self.player_sprite.center_x = self.camera.viewport_width // 2

        # Track Player Motion
        self.x_direction = 0
        self.y_direction = 0

    def pan_camera_to_player(self):
        # This will center the camera on the player.
        target_x = self.player_sprite.center_x
        target_y = self.player_sprite.center_y

        # This ensures the background is always at least partially visible.
        if 0.0 > target_x:
            target_x = 0.0
        elif target_x > self.background.size[0]:
            target_x = self.background.size[0]

        if 0.0 > target_y:
            target_y = 0.0
        elif target_y > self.background.size[1]:
            target_y = self.background.size[1]
        self.camera.position = arcade.math.lerp_2d(
            self.camera.position,
            (target_x, target_y),
            CAMERA_SPEED,
        )

    def on_update(self, delta_time: float):
        new_position = (
            self.player_sprite.center_x + self.x_direction * delta_time,
            self.player_sprite.center_y + self.y_direction * delta_time,
        )
        self.player_sprite.position = new_position

        self.pan_camera_to_player()

    def on_draw(self):
        self.clear()
        with self.camera.activate():
            self.background.draw()
            arcade.draw_sprite(self.player_sprite)

    def on_key_press(self, symbol: int, modifiers: int):
        if symbol in (arcade.key.LEFT, arcade.key.A):
            self.x_direction -= PLAYER_SPEED
        elif symbol in (arcade.key.RIGHT, arcade.key.D):
            self.x_direction += PLAYER_SPEED
        elif symbol in (arcade.key.DOWN, arcade.key.S):
            self.y_direction -= PLAYER_SPEED
        elif symbol in (arcade.key.UP, arcade.key.W):
            self.y_direction += PLAYER_SPEED
        # Close the window if the user presses the escape key
        elif symbol == arcade.key.ESCAPE:
            self.window.close()

    def on_key_release(self, symbol: int, modifiers: int):
        if symbol in (arcade.key.LEFT, arcade.key.A):
            self.x_direction += PLAYER_SPEED
        elif symbol in (arcade.key.RIGHT, arcade.key.D):
            self.x_direction -= PLAYER_SPEED
        elif symbol in (arcade.key.DOWN, arcade.key.S):
            self.y_direction += PLAYER_SPEED
        elif symbol in (arcade.key.UP, arcade.key.W):
            self.y_direction -= PLAYER_SPEED

    def on_resize(self, width: int, height: int):
        super().on_resize(width, height)
        self.camera.match_window()


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE, resizable=True)

    # Create and setup the GameView
    game = GameView()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



from pyglet.math import Vec3

from arcade.camera.data_types import CameraData
from arcade.math import quaternion_rotation

__all__ = ("rotate_around_forward", "rotate_around_up", "rotate_around_right")


def rotate_around_forward(data: CameraData, angle: float) -> tuple[float, float, float]:
    """
    Rotate the CameraData up vector around the CameraData forward vector, perfect
    for rotating the screen.

    This rotation will be around (0.0, 0.0) of the camera projection.
    If that is not the center of the screen this method may appear erroneous.
    Uses arcade.camera.controllers.quaternion_rotation internally.

    Args:
        data:
            The camera data to modify. The data's up vector is rotated around
            its forward vector
        angle:
            The angle in degrees to rotate clockwise by
    """
    return quaternion_rotation(data.forward, data.up, angle)


def rotate_around_up(data: CameraData, angle: float) -> tuple[float, float, float]:
    """
    Rotate the CameraData forward vector around the CameraData up vector.
    Generally only useful in 3D games.
    Uses arcade.camera.controllers.quaternion_rotation internally.

    Args:
        data:
            The camera data to modify. The data's forward vector is rotated
            around its up vector
        angle:
            The angle in degrees to rotate clockwise by
    """
    return quaternion_rotation(data.up, data.forward, angle)


def rotate_around_right(
    data: CameraData, angle: float
) -> tuple[tuple[float, float, float], tuple[float, float, float]]:
    """
    Rotate both the CameraData's forward vector and up vector around a calculated
    right vector. Generally only useful in 3D games.
    Uses arcade.camera.controllers.quaternion_rotation internally.

    Args:
        data:
            The camera data to modify. The data's forward vector is rotated
            around its up vector
        angle:
            The angle in degrees to rotate clockwise by
    """
    _forward = Vec3(data.forward[0], data.forward[1], data.forward[2])
    _up = Vec3(data.up[0], data.up[1], data.up[2])
    _crossed_vec = _forward.cross(_up)
    _right: tuple[float, float, float] = (_crossed_vec.x, _crossed_vec.y, _crossed_vec.z)
    new_forward = quaternion_rotation(_right, data.forward, angle)
    new_up = quaternion_rotation(_right, data.up, angle)
    return new_up, new_forward



"""
Drawing Gradients

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.gradients
"""
import arcade
from arcade import shape_list

# Do the math to figure out our screen dimensions
WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Gradients Example"


class GameView(arcade.View):
    """
    Main application class.
    """

    def __init__(self):
        """
        Set up the application.
        """

        super().__init__()

        self.background_color = arcade.color.BLACK

        self.shapes = shape_list.ShapeElementList()

        # This is a large rectangle that fills the whole
        # background. The gradient goes between the two colors
        # top to bottom.
        color1 = (215, 214, 165)
        color2 = (219, 166, 123)
        points = (
            (0, 0),
            (WINDOW_WIDTH, 0),
            (WINDOW_WIDTH, WINDOW_HEIGHT),
            (0, WINDOW_HEIGHT),
        )
        colors = (color1, color1, color2, color2)
        rect = shape_list.create_rectangle_filled_with_colors(points, colors)
        self.shapes.append(rect)

        # Another rectangle, but in this case the color doesn't change. Just the
        # transparency. This time it goes from left to right.
        color1 = (165, 92, 85, 255)
        color2 = (165, 92, 85, 0)
        points = (
            (100, 100),
            (WINDOW_WIDTH - 100, 100),
            (WINDOW_WIDTH - 100, 300),
            (100, 300),
        )
        colors = (color2, color1, color1, color2)
        rect = shape_list.create_rectangle_filled_with_colors(points, colors)
        self.shapes.append(rect)

        # Two lines
        color1 = (7, 67, 88)
        color2 = (69, 137, 133)
        points = (
            (100, 400),
            (WINDOW_WIDTH - 100, 400),
            (WINDOW_WIDTH - 100, 500),
            (100, 500),
        )
        colors = [color2, color1, color2, color1]
        shape = shape_list.create_lines_with_colors(points, colors, line_width=5)
        self.shapes.append(shape)

        # Triangle
        color1 = (215, 214, 165)
        color2 = (219, 166, 123)
        color3 = (165, 92, 85)
        points = (
            (WINDOW_WIDTH // 2, 500),
            (WINDOW_WIDTH // 2 - 100, 400),
            (WINDOW_WIDTH // 2 + 100, 400),
        )
        colors = (color1, color2, color3)
        shape = shape_list.create_triangles_filled_with_colors(points, colors)
        self.shapes.append(shape)

        # Ellipse, gradient between center and outside
        color1 = (69, 137, 133, 127)
        color2 = (7, 67, 88, 127)
        shape = shape_list.create_ellipse_filled_with_colors(
            WINDOW_WIDTH // 2, 350, 50, 50,
            inside_color=color1, outside_color=color2,
        )
        self.shapes.append(shape)

    def on_draw(self):
        """
        Render the screen.
        """

        # This command has to happen before we start drawing
        self.clear()
        self.shapes.draw()
        # arcade.draw_rectangle_filled(500, 500, 50, 50, (255, 0, 0, 127))


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create the GameView
    game = GameView()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



import warnings
from copy import deepcopy
from dataclasses import dataclass
from typing import Literal

import pyglet
from pyglet.event import EVENT_HANDLED, EVENT_UNHANDLED
from pyglet.text.caret import Caret
from pyglet.text.document import AbstractDocument
from typing_extensions import override

import arcade
from arcade import uicolor
from arcade.gui.events import (
    UIEvent,
    UIMouseDragEvent,
    UIMouseEvent,
    UIMousePressEvent,
    UIMouseScrollEvent,
    UIOnChangeEvent,
    UIOnClickEvent,
    UITextInputEvent,
    UITextMotionEvent,
    UITextMotionSelectEvent,
)
from arcade.gui.property import Property, bind
from arcade.gui.style import UIStyleBase, UIStyledWidget
from arcade.gui.surface import Surface
from arcade.gui.widgets import UIInteractiveWidget, UIWidget
from arcade.gui.widgets.layout import UIAnchorLayout
from arcade.text import FontNameOrNames
from arcade.types import LBWH, RGBA255, Color, RGBOrA255


class UILabel(UIWidget):
    """A simple text label. This widget is meant to display user instructions or
    information. This label supports multiline text.

    If you want to make a scrollable viewing text box, use a
    :py:class:`~arcade.gui.UITextArea`.

    By default, a label will fit its initial content. If the text is changed use
    :py:meth:`~arcade.gui.UILabel.fit_content` to adjust the size.

    If the text changes frequently, ensure to set a background color or texture, which will
    prevent a full rendering of the whole UI and only render the label itself.

    Args:
        text: Text displayed on the label.
        x: x position (default anchor is bottom-left).
        y: y position (default anchor is bottom-left).
        width: Width of the label. Defaults to text width if not
            specified. See
            :py:meth:`~pyglet.text.layout.TextLayout.content_width`.
        height: Height of the label. Defaults to text height if not
            specified. See
            :py:meth:`~pyglet.text.layout.TextLayout.content_height`.
        font_name: A list of fonts to use. Arcade will start at the
            beginning of the tuple and keep trying to load fonts until
            success.
        font_size: Font size of font.
        text_color: Color of the text.
        bold: If enabled, the label's text will be in a **bold** style.
        italic: If enabled, the label's text will be in an *italic*
            style.
        stretch: Stretch font style.
        align: Horizontal alignment of text on a line. This only applies
            if a width is supplied. Valid options include ``"left"``,
            ``"center"`` or ``"right"``.
        dpi: Resolution of the fonts in the layout. Defaults to 96.
        multiline: If enabled, a ``\\n`` will start a new line. Changing
            text or font will require a manual call of
            :py:meth:`~arcade.gui.UILabel.fit_content` to prevent text
            line wrap.
        size_hint: A tuple of floats between 0 and 1 defining the amount
            of space of the parent should be requested. Default (0, 0)
            which fits the content.
        size_hint_max: Maximum size hint width and height in pixel.
    """

    ADAPTIVE_MULTILINE_WIDTH = 999999

    def __init__(
        self,
        text: str = "",
        *,
        x: float = 0,
        y: float = 0,
        width: float | None = None,
        height: float | None = None,
        font_name=("calibri", "arial"),
        font_size: float = 12,
        text_color: RGBOrA255 = arcade.color.WHITE,
        bold: str | bool = False,
        italic=False,
        align="left",
        multiline: bool = False,
        size_hint=(0, 0),
        size_hint_max=None,
        **kwargs,
    ):
        # If multiline is enabled and no width is given, we need to fit the
        # size to the text. This is done by setting the width to a very
        # large value and then fitting the size.
        adaptive_multiline = False
        if multiline and not width:
            width = self.ADAPTIVE_MULTILINE_WIDTH
            adaptive_multiline = True

        # Use Arcade Text wrapper of pyglet.Label for text rendering
        self._label = arcade.Text(
            x=0,
            y=0,
            text=text,
            font_name=font_name,
            font_size=font_size,
            color=text_color,
            width=int(width) if width else None,
            bold=bold,
            italic=italic,
            align=align,
            anchor_y="bottom",  # Position text bottom left to fit into scissor area
            multiline=multiline,
            **kwargs,
        )
        self._strong_background = True

        if adaptive_multiline:
            # +1 is required to prevent line wrap
            width = self._label.content_width + 1

        super().__init__(
            x=x,
            y=y,
            width=width or self._label.content_width,
            height=height or self._label.content_height,
            size_hint=size_hint,
            size_hint_max=size_hint_max,
            **kwargs,
        )

        # Set the label size. If the width or height was given because border
        # and padding can only be applied later, we can avoid ``fit_content``
        # and set with and height separately.
        if width:
            self._label.width = int(width)
        if height:
            self._label.height = int(height)

        bind(self, "rect", self._update_label)

        # update size hint when border or padding changes
        bind(self, "_border_width", self._update_size_hint_min)
        bind(self, "_padding_left", self._update_size_hint_min)
        bind(self, "_padding_right", self._update_size_hint_min)
        bind(self, "_padding_top", self._update_size_hint_min)
        bind(self, "_padding_bottom", self._update_size_hint_min)

        self._update_size_hint_min()

    def fit_content(self):
        """Manually set the width and height of the label to contain the whole text.
        Based on the size_hint_min.

        If multiline is enabled, the width will be calculated based on longest line of the text.
        And size_hint_min will be updated.
        """

        if self.multiline:
            self._label.width = self.ADAPTIVE_MULTILINE_WIDTH
            self._update_size_hint_min()

        min_width, min_height = self.size_hint_min or (1, 1)
        self.rect = self.rect.resize(
            width=min_width,
            height=min_height,
        )
        # rect changes to trigger resizing of the _label automatically

    @property
    def text(self):
        """Text of the label."""
        return self._label.text

    @text.setter
    def text(self, value):
        """Update text of the label.

        This triggers a full render to ensure that previous text is cleared out.
        """

        if self._label.text != value:
            self._label.text = value
            self._update_size_hint_min()

            if self._bg_color or self._bg_tex:
                self.trigger_render()
            else:
                self.trigger_full_render()

    @property
    def font_name(self) -> FontNameOrNames:
        """Font name of the label. Use :py:meth:`~arcade.gui.UILabel.update_font` to change."""
        return self._label.font_name

    @property
    def font_size(self) -> float:
        """Font size of the label. Use :py:meth:`~arcade.gui.UILabel.update_font` to change."""
        return self._label.font_size

    @property
    def font_color(self) -> Color:
        """Font color of the label. Use :py:meth:`~arcade.gui.UILabel.update_font` to change."""
        return self._label.color

    @property
    def bold(self) -> bool | str:
        """Return if the label is in bold style.
        Use :py:meth:`~arcade.gui.UILabel.update_font` to change."""
        return self._label.bold

    @property
    def italic(self) -> bool | str:
        """Return if the label is in italic style.
        Use :py:meth:`~arcade.gui.UILabel.update_font` to change."""
        return self._label.italic

    def _update_label(self):
        """Update the position and size of the label.

        So it fits into the content area of the widget.
        Should always be called after the content area changed.
        """
        # Update Pyglet label size
        label = self._label
        layout_size = label.width, label.height

        if layout_size != self.content_size or label.position != (0, 0):
            label.position = 0, 0, 0  # label always drawn in scissor box
            label.width = int(self.content_width)
            label.height = int(self.content_height)

    def _update_size_hint_min(self):
        """Update the minimum size hint based on the label content size."""
        min_width = self._label.content_width + 1  # +1 required to prevent line wrap
        min_width += self._padding_left + self._padding_right + 2 * self._border_width

        min_height = self._label.content_height
        min_height += self._padding_top + self._padding_bottom + 2 * self._border_width

        self.size_hint_min = (min_width, min_height)

    def update_font(
        self,
        font_name: FontNameOrNames | None = None,
        font_size: float | None = None,
        font_color: Color | None = None,
        bold: bool | str | None = None,
        italic: bool | None = None,
    ):
        """Update font of the label.

        Args:
            font_name: A list of fonts to use. Arcade will start at the
                beginning of the tuple and keep trying to load fonts until
                success.
            font_size: Font size of font.
            font_color: Color of the text.
            bold: May be any value in :py:obj:`pyglet.text.Weight`,
                ``True`` (converts to ``"bold"``), or ``False``
                (converts to ``"regular"``).
            italic: If enabled, the label's text will be in an *italic*
        """
        font_name = font_name or self._label.font_name
        font_size = font_size or self._label.font_size
        font_color = font_color or self._label.color
        font_bold = bold if bold is not None else self._label.bold
        font_italic = italic if italic is not None else self._label.italic

        # ensure type of font_color, label will allways be a color
        font_color = Color.from_iterable(font_color)

        # Check if values actually changed, if then update and trigger render
        font_name_changed = self._label.font_name != font_name
        font_size_changed = self._label.font_size != font_size
        font_color_changed = self._label.color != font_color
        font_bold_changed = self._label.bold != font_bold
        font_italic_changed = self._label.italic != font_italic
        if (
            font_name_changed
            or font_size_changed
            or font_color_changed
            or font_bold_changed
            or font_italic_changed
        ):
            with self._label:
                self._label.font_name = font_name
                self._label.font_size = font_size
                self._label.color = font_color
                self._label.bold = font_bold
                self._label.italic = font_italic
            self._update_size_hint_min()

            # Optimised render behaviour
            if self._bg_color or self._bg_tex:
                self.trigger_render()
            else:
                self.trigger_full_render()

    @property
    def multiline(self) -> bool:
        """Return if the label is in multiline mode."""
        return self._label.multiline

    def do_render(self, surface: Surface):
        """Render the label via py:class:`~arcade.Text`."""
        self.prepare_render(surface)

        # pyglet rendering automatically applied by arcade.Text
        self._label.draw()


class UITextWidget(UIAnchorLayout):
    """Adds the ability to add text to a widget.
    Use this to create subclass widgets, which have text.

    The text can be placed within the widget using
    :py:class:`~arcade.gui.UIAnchorLayout` parameters with
    :py:meth:`~arcade.gui.UITextWidget.place_text`.

    The widget holds reference to one primary :py:class:`~arcade.gui.UILabel`, which is placed in
    the widget's layout. This label can be accessed
    via :py:attr:`~arcade.gui.UITextWidget.ui_label`.

    To change font, font size, or text color, use py:meth:`~arcade.gui.UILabel.update_font`.

    Args:
        text: Text displayed on the label.
        multiline: If enabled, a ``\\n`` will start a new line.
        **kwargs: passed to :py:class:`~arcade.gui.UIWidget`.

    """

    def __init__(self, *, text: str, multiline: bool = False, **kwargs):
        super().__init__(text=text, **kwargs)
        self._restrict_child_size = True
        self._label = UILabel(
            text=text, multiline=multiline
        )  # UILabel supports width=None for multiline
        self.add(self._label)

    def place_text(
        self,
        anchor_x: str | None = None,
        align_x: float = 0,
        anchor_y: str | None = None,
        align_y: float = 0,
        **kwargs,
    ) -> UILabel:
        """Place widget's text within the widget using
        :py:class:`~arcade.gui.UIAnchorLayout` parameters.

        Args:
            anchor_x: Horizontal anchor. Valid options are ``left``,
                ``right``, and ``center``.
            align_x: Offset or padding for the horizontal anchor.
            anchor_y: Vertical anchor. Valid options are ``top``,
                ``center``, and ``bottom``.
            align_y: Offset or padding for the vertical anchor.
            **kwargs: Additional keyword arguments passed to the layout function.
        """
        self.remove(self._label)
        return self.add(
            child=self._label,
            anchor_x=anchor_x,
            align_x=align_x,
            anchor_y=anchor_y,
            align_y=align_y,
            **kwargs,
        )

    @property
    def text(self):
        """Text of the widget. Modifying this repeatedly will cause significant
        lag; calculating glyph position is very expensive.
        """
        return self.ui_label.text

    @text.setter
    def text(self, value):
        self.ui_label.text = value
        self.trigger_render()

    @property
    def multiline(self):
        """Get or set the multiline mode.

        Newline characters (``"\\n"``) will only be honored when this is set to ``True``.
        If you want a scrollable text widget, please use :py:class:`~arcade.gui.UITextArea`
        instead.
        """
        return self.ui_label.multiline

    @property
    def ui_label(self) -> UILabel:
        """Internal py:class:`~arcade.gui.UILabel` used for rendering the text."""
        return self._label


@dataclass
class UIInputTextStyle(UIStyleBase):
    """Used to style the UITextWidget for different states. Below is its use case.

    .. code:: py

        button = UIInputText(style={"normal": UIInputText.UIStyle(...),})

    Args:
        bg: Background color.
        border: Border color.
        border_width: Width of the border.

    """

    bg: RGBA255 | None = None
    border: RGBA255 | None = uicolor.WHITE
    border_width: int = 2


class UIInputText(UIStyledWidget[UIInputTextStyle], UIInteractiveWidget):
    """An input field the user can type text into.

    This is useful in returning
    string input from the user. A caret is displayed, which the user can move
    around with a mouse or keyboard.

    A mouse drag selects text, a mouse press moves the caret, and keys can move
    around the caret. Arcade confirms that the field is active before allowing
    users to type, so it is okay to have multiple of these.

    By default, a border is drawn around the input field.

    The widget emits a :py:class:`~arcade.gui.UIOnChangeEvent` event when the text changes.

    Args:
        x: x position (default anchor is bottom-left).
        y: y position (default anchor is bottom-left).
        width: Width of the text field.
        height: Height of the text field.
        text: Initial text displayed. This can be modified later
            programmatically or by the user's interaction with the
            caret.
        font_name: A list of fonts to use. Arcade will start at the
            beginning of the tuple and keep trying to load fonts until
            success.
        font_size: Font size of font.
        text_color: Color of the text.
        multiline: If enabled, a ``\\n`` will start a new line. A
            :py:class:`~arcade.gui.UITextWidget`  ``multiline`` of True
            is the same thing as a :py:class:`~arcade.gui.UITextArea`.
        caret_color: An RGBA or RGB color for the caret with each
            channel between 0 and 255, inclusive.
        size_hint: A tuple of floats between 0 and 1 defining the amount
            of space of the parent should be requested.
        size_hint_min: Minimum size hint width and height in pixel.
        size_hint_max: Maximum size hint width and height in pixel.
        **kwargs: passed to :py:class:`~arcade.gui.UIWidget`.

    """

    # Move layout one pixel into the scissor box so the caret is also shown at
    # position 0.
    LAYOUT_OFFSET = 1

    # Style
    UIStyle = UIInputTextStyle

    DEFAULT_STYLE = {
        "normal": UIStyle(),
        "hover": UIStyle(
            border=uicolor.WHITE_CLOUDS,
        ),
        "press": UIStyle(
            border=uicolor.WHITE_SILVER,
        ),
        "disabled": UIStyle(
            bg=uicolor.WHITE_SILVER,
        ),
        "invalid": UIStyle(
            bg=uicolor.RED_ALIZARIN.replace(a=42),
            border=uicolor.RED_ALIZARIN,
        ),
    }

    # Properties
    invalid = Property(False)

    def __init__(
        self,
        *,
        x: float = 0,
        y: float = 0,
        width: float = 100,
        height: float = 25,  # required height for font size 12 + border width 1
        text: str = "",
        font_name=("Arial",),
        font_size: float = 12,
        text_color: RGBOrA255 = arcade.color.WHITE,
        multiline=False,
        caret_color: RGBOrA255 = arcade.color.WHITE,
        border_color: Color | None = arcade.color.WHITE,
        border_width: int = 2,
        size_hint=None,
        size_hint_min=None,
        size_hint_max=None,
        style: dict[str, UIInputTextStyle] | None = None,
        **kwargs,
    ):
        if border_color != arcade.color.WHITE or border_width != 2:
            warnings.warn(
                "UIInputText is now a UIStyledWidget. "
                "Use the style dict to set the border color and width.",
                DeprecationWarning,
                stacklevel=1,
            )

            # adjusting style to set border color and width
            style = style or UIInputText.DEFAULT_STYLE
            style = deepcopy(style)

            style["normal"].border = border_color
            style["normal"].border_width = border_width

        super().__init__(
            x=x,
            y=y,
            width=width,
            height=height,
            size_hint=size_hint,
            size_hint_min=size_hint_min,
            size_hint_max=size_hint_max,
            style=style or UIInputText.DEFAULT_STYLE,
            **kwargs,
        )

        self._active = False
        self._text_color = Color.from_iterable(text_color)

        self.doc: AbstractDocument = pyglet.text.decode_text(text)
        self.doc.set_style(
            0,
            len(text),
            dict(font_name=font_name, font_size=font_size, color=self._text_color),
        )

        self.layout = pyglet.text.layout.IncrementalTextLayout(
            self.doc,
            x=0 + self.LAYOUT_OFFSET,
            y=0,
            z=0.0,  # Position
            width=int(width - self.LAYOUT_OFFSET),
            height=int(height),  # Size
            multiline=multiline,
        )
        self.caret = Caret(self.layout, color=Color.from_iterable(caret_color))
        self.caret.visible = False

        self._blink_state = self._get_caret_blink_state()

        self.register_event_type("on_change")

        bind(self, "hovered", self._apply_style)
        bind(self, "pressed", self._apply_style)
        bind(self, "invalid", self._apply_style)
        bind(self, "disabled", self._apply_style)

        # initial style application
        self._apply_style()

    def _apply_style(self):
        style = self.get_current_style()

        self.with_background(
            color=Color.from_iterable(style.bg) if style.bg else None,
        )
        self.with_border(
            color=Color.from_iterable(style.border) if style.border else None,
            width=style.border_width,
        )
        self.trigger_full_render()

    @override
    def get_current_state(self) -> str:
        """Get the current state of the slider.

        Returns:
            ""normal"", ""hover"", ""press"" or ""disabled"".
        """
        if self.disabled:
            return "disabled"
        elif self.pressed:
            return "press"
        elif self.hovered:
            return "hover"
        elif self.invalid:
            return "invalid"
        else:
            return "normal"

    def _get_caret_blink_state(self):
        """Check whether or not the caret is currently blinking or not."""
        return self.caret.visible and self._active and self.caret._blink_visible

    @override
    def on_update(self, dt):
        """Update the caret blink state."""
        # Only trigger render if blinking state changed
        current_state = self._get_caret_blink_state()
        if self._blink_state != current_state:
            self._blink_state = current_state
            self.trigger_full_render()

    def on_click(self, event: UIOnClickEvent):
        self.activate()

    @override
    def on_event(self, event: UIEvent) -> bool | None:
        """Handle events for the text input field.

        Text input is only active when the user clicks on the input field."""
        # If active check to deactivate
        if self._active and isinstance(event, UIMousePressEvent):
            if self.rect.point_in_rect(event.pos):
                x = int(event.x - self.left - self.LAYOUT_OFFSET)
                y = int(event.y - self.bottom)
                self.caret.on_mouse_press(x, y, event.button, event.modifiers)
            else:
                self.deactivate()
                # return unhandled to allow other widgets to activate
                return EVENT_UNHANDLED

        # If active pass all non press events to caret
        if self._active:
            old_text = self.text
            # Act on events if active
            if isinstance(event, UITextInputEvent):
                self.caret.on_text(event.text)
                self.trigger_full_render()
            elif isinstance(event, UITextMotionEvent):
                self.caret.on_text_motion(event.motion)
                self.trigger_full_render()
            elif isinstance(event, UITextMotionSelectEvent):
                self.caret.on_text_motion_select(event.selection)
                self.trigger_full_render()

            if isinstance(event, UIMouseEvent) and self.rect.point_in_rect(event.pos):
                x = int(event.x - self.left - self.LAYOUT_OFFSET)
                y = int(event.y - self.bottom)
                if isinstance(event, UIMouseDragEvent):
                    self.caret.on_mouse_drag(
                        x, y, event.dx, event.dy, event.buttons, event.modifiers
                    )
                    self.trigger_full_render()
                elif isinstance(event, UIMouseScrollEvent):
                    self.caret.on_mouse_scroll(x, y, event.scroll_x, event.scroll_y)
                    self.trigger_full_render()

            if old_text != self.text:
                self.dispatch_event("on_change", UIOnChangeEvent(self, old_text, self.text))

        return super().on_event(event)

    @property
    def active(self) -> bool:
        """Return if the text input field is active.

        An active text input field will show a caret and accept text input."""
        return self._active

    def activate(self):
        """Programmatically activate the text input field."""
        if self._active:
            return

        self._active = True
        self.trigger_full_render()
        self.caret.on_activate()
        self.caret.position = len(self.doc.text)

    def deactivate(self):
        """Programmatically deactivate the text input field."""

        if not self._active:
            return

        self._active = False
        self.trigger_full_render()
        self.caret.on_deactivate()

    def _update_layout(self):
        # Update Pyglet layout size
        layout = self.layout
        layout_size = layout.width - self.LAYOUT_OFFSET, layout.height

        if layout_size != self.content_size:
            layout.begin_update()
            layout.width = int(self.content_width - self.LAYOUT_OFFSET)
            layout.height = int(self.content_height)

            # should not be required, but the caret does not show up on first click without text
            layout.x = self.LAYOUT_OFFSET
            layout.y = 0
            layout.end_update()

    @property
    def text(self):
        """Text of the input field."""
        return self.doc.text

    @text.setter
    def text(self, value):
        if value != self.doc.text:
            old_text = self.doc.text
            self.doc.text = value
            self.dispatch_event("on_change", UIOnChangeEvent(self, old_text, self.text))

            # if bg color or texture is set, render this widget only
            if self._bg_color or self._bg_tex:
                self.trigger_render()
            else:
                self.trigger_full_render()

    @override
    def do_render(self, surface: Surface):
        """Render the text input field."""
        self._update_layout()
        self.prepare_render(surface)

        self.layout.draw()

    def on_change(self, event: UIOnChangeEvent):
        """Event handler for text change."""
        pass


class UITextArea(UIWidget):
    """A text area that allows users to view large documents of text by scrolling
    the mouse.

    Args:
        x: x position (default anchor is bottom-left).
        y: y position (default anchor is bottom-left).
        width: Width of the text area.
        height: Height of the text area.
        text: Initial text displayed.
        font_name: A list of fonts to use. Arcade will start at the
            beginning of the tuple and keep trying to load fonts until
            success.
        font_size: Font size of font.
        text_color: Color of the text.
        bold: If enabled, the label's text will be in a **bold** style.
        italic: If enabled, the label's text will be in an *italic*
        multiline: If enabled, a ``\\n`` will start a new line.
        scroll_speed: Speed of mouse scrolling.
        size_hint: A tuple of floats between 0 and 1 defining the amount
            of space of the parent should be requested.
        size_hint_min: Minimum size hint width and height in pixel.
        size_hint_max: Maximum size hint width and height in pixel.
        document_mode: Mode of the document. Can be "PLAIN", "ATTRIBUTED", or "HTML".
            PLAIN will decode the text as plain text, ATTRIBUTED and HTML will
            decode the text as pyglet documents here
            https://pyglet.readthedocs.io/en/latest/programming_guide/text.html
        **kwargs: passed to :py:class:`~arcade.gui.UIWidget`.
    """

    def __init__(
        self,
        *,
        x: float = 0,
        y: float = 0,
        width: float = 400,
        height: float = 40,
        text: str = "",
        font_name=("arial", "calibri"),
        font_size: float = 12,
        bold=False,
        italic=False,
        text_color: RGBA255 = arcade.color.WHITE,
        multiline: bool = True,
        scroll_speed: float | None = None,
        size_hint=None,
        size_hint_min=None,
        size_hint_max=None,
        document_mode: Literal["PLAIN", "ATTRIBUTED", "HTML"] = "PLAIN",
        **kwargs,
    ):
        super().__init__(
            x=x,
            y=y,
            width=width,
            height=height,
            size_hint=size_hint,
            size_hint_min=size_hint_min,
            size_hint_max=size_hint_max,
            **kwargs,
        )

        # Set how fast the mouse scroll wheel will scroll text in the pane.
        # Measured in pixels per 'click'
        self.scroll_speed = scroll_speed if scroll_speed is not None else font_size

        self.doc: AbstractDocument
        if document_mode == "PLAIN":
            self.doc = pyglet.text.decode_text(text)
        elif document_mode == "ATTRIBUTED":
            self.doc = pyglet.text.decode_attributed(text)
        elif document_mode == "HTML":
            self.doc = pyglet.text.decode_html(text)

        self.doc.set_style(
            0,
            len(text),
            dict(
                font_name=font_name,
                font_size=font_size,
                color=Color.from_iterable(text_color),
                bold=bold,
                italic=italic,
            ),
        )

        self.layout = pyglet.text.layout.ScrollableTextLayout(
            self.doc,
            width=int(self.content_width),
            height=int(self.content_height),
            multiline=multiline,
        )

        # bind(self, "rect", self._update_layout)

    def fit_content(self):
        """Set the width and height of the text area to contain the whole text."""
        self.rect = LBWH(
            self.left,
            self.bottom,
            self.layout.content_width,
            self.layout.content_height,
        )

    @property
    def text(self):
        """Text of the text area."""
        return self.doc.text

    @text.setter
    def text(self, value):
        self.doc.text = value
        self.trigger_render()

    def _update_layout(self):
        # Update Pyglet layout size
        layout = self.layout

        # Convert from local float coords to ints to avoid jitter
        # since pyglet imposes int-only coordinates as of pyglet 2.0
        content_width, content_height = map(int, self.content_size)
        if content_width != layout.width or content_height != layout.height:
            layout.begin_update()
            layout.width = content_width
            layout.height = content_height
            layout.end_update()

    @override
    def do_render(self, surface: Surface):
        """Render the text area."""
        self._update_layout()
        self.prepare_render(surface)
        self.layout.draw()

    @override
    def on_event(self, event: UIEvent) -> bool | None:
        """Handle scrolling of the widget."""
        if isinstance(event, UIMouseScrollEvent):
            if self.rect.point_in_rect(event.pos):
                self.layout.view_y = round(self.layout.view_y + event.scroll_y * self.scroll_speed)
                self.trigger_full_render()

        if super().on_event(event):
            return EVENT_HANDLED

        return EVENT_UNHANDLED



"""Color-related types, aliases, and constants.

This module does not contain pre-defined color values. For pre-made
named color values, please see the following:

.. list-table::
   :header-rows: 1

   * - Module
     - Contents

   * - :py:mod:`arcade.color`
     - A set of pre-defined :py:class`.Color` constants.

   * - :py:mod:`arcade.csscolor`
     - The `CSS named colors <https://developer.mozilla.org/en-US/docs/Web/CSS/named-color>`_
       as :py:class:`.Color` constants.

"""

from __future__ import annotations

import random
from collections.abc import Iterable
from typing import Final, TypeVar

from typing_extensions import Self

from arcade.exceptions import ByteRangeError, IntOutsideRangeError, NormalizedRangeError

__all__ = (
    "Color",
    "RGB",
    "RGBA",
    "RGB255",
    "RGBA255",
    "RGBNormalized",
    "RGBANormalized",
    "RGBOrA",
    "RGBOrA255",
    "RGBOrANormalized",
    "MASK_RGBA_R",
    "MASK_RGBA_G",
    "MASK_RGBA_B",
    "MASK_RGBA_A",
    "MASK_RGB_R",
    "MASK_RGB_G",
    "MASK_RGB_B",
    "MAX_UINT24",
    "MAX_UINT32",
)


# Helpful color-related constants for bit masking
MAX_UINT24: Final[int] = 0xFFFFFF
MAX_UINT32: Final[int] = 0xFFFFFFFF
MASK_RGBA_R: Final[int] = 0xFF000000
MASK_RGBA_G: Final[int] = 0x00FF0000
MASK_RGBA_B: Final[int] = 0x0000FF00
MASK_RGBA_A: Final[int] = 0x000000FF
MASK_RGB_R: Final[int] = 0xFF0000
MASK_RGB_G: Final[int] = 0x00FF00
MASK_RGB_B: Final[int] = 0x0000FF


# Color type aliases.
ChannelType = TypeVar("ChannelType", int, float)

# Generic color aliases
RGB = tuple[ChannelType, ChannelType, ChannelType]
RGBA = tuple[ChannelType, ChannelType, ChannelType, ChannelType]
RGBOrA = RGB[ChannelType] | RGBA[ChannelType]

# Specific color aliases
RGB255 = RGB[int]
RGBA255 = RGBA[int]
RGBNormalized = RGB[float]
RGBANormalized = RGBA[float]
RGBOrA255 = RGBOrA[int]
RGBOrANormalized = RGBOrA[float]


class Color(RGBA255):
    """An RGBA color as a :py:class:`tuple` subclass.

    .. code-block:: python

        # The alpha channel value defaults to 255
        >>> from arcade.types import Color
        >>> Color(255, 0, 0)
        Color(r=255, g=0, b=0, a=255)

    If you prefer specifying color with another format, the class also
    provides number of helper methods for the most common RGB and RGBA
    formats:

    * :py:meth:`.from_hex_string`
    * :py:meth:`.from_normalized`
    * :py:meth:`.from_uint24`
    * :py:meth:`.from_uint32`
    * :py:meth:`.from_iterable`

    Regardless of the source format, all color channels must be between
    0 and 255, inclusive. If any channel is outside this range, creation
    will fail with a :py:class:`~arcade.utils.ByteRangeError`, which is a
    type of :py:class:`ValueError`.

    .. _colour: https://pypi.org/project/colour/

    .. note:: This class does not currently support HSV or other color spaces.

              If you need these, you may want to try the following:

              * Python's built-in :py:mod:`colorsys` module
              * The `colour`_ package

    Args:
        r: the red channel of the color, between 0 and 255, inclusive
        g: the green channel of the color, between 0 and 255, inclusive
        b: the blue channel of the color, between 0 and 255, inclusive
        a: the alpha or transparency channel of the color, between 0 and 255, inclusive
    """

    __slots__ = ()

    def __new__(cls, r: int, g: int, b: int, a: int = 255):
        if not 0 <= r <= 255:
            raise ByteRangeError("r", r)

        if not 0 <= g <= 255:
            raise ByteRangeError("g", g)

        if not 0 <= g <= 255:
            raise ByteRangeError("b", b)

        if not 0 <= a <= 255:
            raise ByteRangeError("a", a)

        # Typechecking is ignored because of a mypy bug involving
        # tuples & super:
        # https://github.com/python/mypy/issues/8541
        return super().__new__(cls, (r, g, b, a))  # type: ignore

    def __getnewargs__(self) -> tuple[int, int, int, int]:
        return self.r, self.g, self.b, self.a

    def __deepcopy__(self, _) -> Self:
        """Allow :py:func:`~copy.deepcopy` to be used with Color"""
        return self.__class__(r=self.r, g=self.g, b=self.b, a=self.a)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(r={self.r}, g={self.g}, b={self.b}, a={self.a})"

    @property
    def r(self) -> int:
        """Get the red value of the :py:class:`Color`.

        It will be between ``0`` and ``255``, inclusive.
        """
        return self[0]

    @property
    def g(self) -> int:
        """Get the green value of the :py:class:`Color`.

        It will be between ``0`` and ``255``, inclusive.
        """
        return self[1]

    @property
    def b(self) -> int:
        """Get the blue value of the :py:class:`Color`.

        It will be between ``0`` and ``255``, inclusive.
        """
        return self[2]

    @property
    def a(self) -> int:
        """Get the alpha value of the :py:class:`Color`.

        It will be between ``0`` and ``255``, inclusive.
        """
        return self[3]

    @property
    def rgb(self) -> tuple[int, int, int]:
        """Return only a color's RGB components.

        This is syntactic sugar for slice indexing as below:

        .. code-block:: python

            >>> from arcade.color import WHITE
            >>> WHITE[:3]
            (255, 255, 255)
            # Equivalent but slower than the above
            >>> (WHITE.r, WHITE.g, WHITE.b)
            (255, 255, 255)

        To reorder the channels as you retrieve them, see
        :meth:`.swizzle`.
        """
        return self[:3]

    @classmethod
    def from_iterable(cls, iterable: Iterable[int]) -> Self:
        """Create a :py:class:`Color` from an iterable of 3 or 4 channel values.

        If an iterable is already a :py:class:`Color` instance,
        it will be returned unchanged. Otherwise, it must unpack as
        3 or 4 :py:class:`int` values between ``0`` and ``255``, inclusive.

        If an iterable has no less than 3 or more than 4 elements,
        this method raises a :py:class:`ValueError`. The function will
        attempt to create a new Color instance. The usual rules apply,
        i.e.: all values must be between 0 and 255, inclusive.

        .. note:: This is a more readable alternative to ``*`` unpacking.

                  If you are an advanced user who needs brevity or
                  higher performance, you can unpack directly into
                  :py:class:`Color`:

                  .. code-block:: python

                     >>> rgb_green_tuple = (0, 255, 0)
                     >>> Color(*rgb_green_tuple, 127)
                     Color(r=0, g=255, b=0, a=127)

        Args:
            iterable: An iterable which unpacks to 3 or 4 elements,
                each between 0 and 255, inclusive.
        """
        if isinstance(iterable, cls):
            return iterable

        # We use unpacking because there isn't a good way of specifying
        # lengths for sequences as of 3.8, our minimum Python version as
        # of March 2023: https://github.com/python/typing/issues/786
        r, g, b, *_a = iterable

        if _a:
            if len(_a) > 1:
                raise ValueError("iterable must unpack to 3 or 4 values")
            a = _a[0]
        else:
            a = 255

        return cls(r, g, b, a=a)

    @property
    def normalized(self) -> RGBANormalized:
        """Convert the :py:class:`Color` to a tuple of 4 normalized floats.

        .. code-block:: python

            >>> arcade.color.WHITE.normalized
            (1.0, 1.0, 1.0, 1.0)

            >>> arcade.color.BLACK.normalized
            (0.0, 0.0, 0.0, 1.0)

            >>> arcade.color.TRANSPARENT_BLACK.normalized
            (0.0, 0.0, 0.0, 0.0)

        """
        return self[0] / 255, self[1] / 255, self[2] / 255, self[3] / 255

    @classmethod
    def from_gray(cls, brightness: int, a: int = 255) -> Self:
        """Create a gray :py:class:`Color` of the given ``brightness``.

        .. code-block:: python

            >>> off_white = Color.from_gray(220)
            >>> print(off_white)
            Color(r=220, g=220, b=220, a=255)

            >>> half_opacity_gray = Color.from_gray(128, 128)
            >>> print(half_opacity_gray)
            Color(r=128, g=128, b=128, a=128)

        Args:
            brightness: How bright the new gray should be
            a: a transparency value, fully opaque by default
        """

        if not 0 <= brightness <= 255:
            raise ByteRangeError("brightness", brightness)

        if not 0 <= a <= 255:
            raise ByteRangeError("a", a)

        return cls(brightness, brightness, brightness, a=a)

    @classmethod
    def from_uint24(cls, color: int, a: int = 255) -> Self:
        """Convert an unsigned 24-bit integer to a :py:class:`Color`.

        .. code-block:: python

            # The alpha channel is assumed to be 255
            >>> Color.from_uint24(0x010203)
            Color(r=1, g=2, b=3, a=255)

            # Specify alpha via the a keyword argument
            >>> Color.from_uint24(0x010203, a=127)
            Color(r=1, g=2, b=3, a=127)

            # The maximum value as decimal
            >>> Color.from_uint24(16777215)
            Color(r=255, g=255, b=255, a=255)

        To convert from an RGBA value as a 32-bit integer, see
        :py:meth:`.from_uint32`.

        Args:
            color:
                a 3-byte :py:class:`int` between ``0`` and ``16777215`` (``0xFFFFFF``)
            a:
                An alpha value to use between 0 and 255, inclusive.
        """

        if not 0 <= color <= MAX_UINT24:
            raise IntOutsideRangeError("color", color, 0, MAX_UINT24)

        if not 0 <= a <= 255:
            raise ByteRangeError("a", a)

        return cls((color & 0xFF0000) >> 16, (color & 0xFF00) >> 8, color & 0xFF, a=a)

    @classmethod
    def from_uint32(cls, color: int) -> Self:
        """Convert an unsigned 32-bit integer to a :py:class:`Color`.

        The four bytes are interpreted as R, G, B, A:

        .. code-block:: python

            >>> Color.from_uint32(0x01020304)
            Color(r=1, g=2, b=3, a=4)

            # The maximum value as a decimal integer
            >>> Color.from_uint32(4294967295)
            Color(r=255, g=255, b=255, a=255)

        To convert from an RGB value as a 24-bit integer, see
        :py:meth:`.from_uint24`.

        Args:
            color:
                An :py:class:`int` between ``0`` and ``4294967295`` (``0xFFFFFFFF``)
        """
        if not 0 <= color <= MAX_UINT32:
            raise IntOutsideRangeError("color", color, 0, MAX_UINT32)

        return cls(
            (color & 0xFF000000) >> 24,
            (color & 0xFF0000) >> 16,
            (color & 0xFF00) >> 8,
            a=(color & 0xFF),
        )

    @classmethod
    def from_normalized(cls, color_normalized: RGBANormalized) -> Self:
        """Convert normalized float RGBA to an RGBA :py:class:`Color`.

        If any input channels aren't normalized (between ``0.0`` and
        ``1.0``), this method will raise a
        :py:class:`~arcade.utils.NormalizedRangeError` you can handle as
        a :py:class:`ValueError`.

        Examples::

            >>> Color.from_normalized((1.0, 0.0, 0.0, 1.0))
            Color(r=255, g=0, b=0, a=255)

            >>> normalized_half_opacity_green = (0.0, 1.0, 0.0, 0.5)
            >>> Color.from_normalized(normalized_half_opacity_green)
            Color(r=0, g=255, b=0, a=127)

        Args:
            color_normalized: A tuple of 4 normalized (``0.0`` to ``1.0``) RGBA values.
        """
        r, g, b, *_a = color_normalized

        if _a:
            if len(_a) > 1:
                raise ValueError("color_normalized must unpack to 3 or 4 values")
            a = _a[0]

            if not 0.0 <= a <= 1.0:
                raise NormalizedRangeError("a", a)

        else:
            a = 1.0

        if not 0.0 <= r <= 1.0:
            raise NormalizedRangeError("r", r)

        if not 0.0 <= g <= 1.0:
            raise NormalizedRangeError("g", g)

        if not 0.0 <= b <= 1.0:
            raise NormalizedRangeError("b", b)

        return cls(int(255 * r), int(255 * g), int(255 * b), a=int(255 * a))

    @classmethod
    def from_hex_string(cls, code: str) -> Self:
        """Create a :py:class:`Color` from a hex code of 3, 4, 6, or 8 digits.

        .. code-block:: python

            # RGB color codes are assumed to have an alpha value of 255
            >>> Color.from_hex_string("#FF00FF")
            Color(r=255, g=0, b=255, a=255)

            # You can use eight-digit RGBA codes to specify alpha
            >>> Color.from_hex_string("#FF007F")
            Color(r=255, g=0, b=255, a=127)

            # For brevity, you can omit the # and use RGB shorthand
            >>> Color.from_hex_string("FFF")
            Color(r=255, g=255, b=255, a=255)

            # Lower case and four-digit RGBA shorthand are also allowed
            >>> Color.from_hex_string("ff0a")
            Color(r=255, g=255, b=0, a=170)

        Aside from the optional leading ``#``, the ``code`` must otherwise
        be a valid CSS hexadecimal color code. It will be processed as
        follows:

        * Any leading ``'#'`` characters will be stripped
        * 3 and 4 digit shorthands are expanded by multiplying each
          digit's value by 16
        * 6 digit RGB hex codes assume 255 as their alpha values
        * 8 digit RGBA hex codes are converted to byte values
          and passed directly to a new :py:class:`Color`
        * All other lengths will raise a :py:class:`ValueError`

        .. _CSS hex color: https://www.w3.org/TR/css-color-4/#hex-notation
        .. _Simple Wiki's Hexadecimal Page: https://simple.wikipedia.org/wiki/Hexadecimal

        To learn more, please see:

        * Python's :py:func:`hex` function and :py:class:`int` type
        * `Simple Wiki's Hexadecimal Page`_
        * The `CSS hex color`_ specification

        Args:
            code:
                A `CSS hex color`_ string which may omit the leading ``#`` character.
        """
        code = code.lstrip("#")

        # This looks unusual, but it matches CSS color code expansion
        # behavior for 3 and 4 digit hex codes.
        if len(code) <= 4:
            code = "".join(char * 2 for char in code)

        if len(code) == 6:
            # full opacity if no alpha specified
            return cls(int(code[:2], 16), int(code[2:4], 16), int(code[4:6], 16), 255)
        elif len(code) == 8:
            return cls(
                int(code[:2], 16), int(code[2:4], 16), int(code[4:6], 16), int(code[6:8], 16)
            )

        raise ValueError(f"Improperly formatted color: '{code}'")

    @classmethod
    def random(
        cls,
        r: int | None = None,
        g: int | None = None,
        b: int | None = None,
        a: int | None = None,
    ) -> Self:
        """Create a :py:class:`Color` by randomizing all unspecified channels.

        All arguments are optional. If you specify a channel's value, it
        will be used in the new color instead of randomizing:

        .. code-block:: python

            # Randomize all channels
            >>> Color.random()
            Color(r=35, g=145, b=4, a=200)

            # Create a random opaque color
            >>> Color.random(a=255)
            Color(r=25, g=99, b=234, a=255)

        Args:
            r: Specify a value for the red channel
            g: Specify a value for the green channel
            b: Specify a value for the blue channel
            a: Specify a value for the alpha channel
        """
        rand = random.randint(0, MAX_UINT32)
        if r is None:
            r = (rand & MASK_RGBA_R) >> 24
        if g is None:
            g = (rand & MASK_RGBA_G) >> 16
        if b is None:
            b = (rand & MASK_RGBA_B) >> 8
        if a is None:
            a = rand & MASK_RGBA_A

        return cls(r, g, b, a)

    def replace(
        self,
        r: int | None = None,
        g: int | None = None,
        b: int | None = None,
        a: int | None = None,
    ) -> Color:
        """Create a :py:class:`Color` with specified values replaced in a predefined color.

        .. code-block:: python

            # Color with alpha with a predefined constant
            >>> arcade.color.BLUE.replace(a = 100)
            Color(r = 0, g = 0, b = 255, a = 100)

        Args:
            r: Specify a value for the red channel
            g: Specify a value for the green channel
            b: Specify a value for the blue channel
            a: Specify a value for the alpha channel
        """
        return Color(
            self.r if r is None else r,
            self.g if g is None else g,
            self.b if b is None else b,
            self.a if a is None else a,
        )

    def swizzle(self, order: str) -> tuple[int, ...]:
        """Get a :py:class:`tuple` of channel values in the given ``order``.

        .. _GLSL's swizzling: https://www.khronos.org/opengl/wiki/Data_Type_(GLSL)#Swizzling

        This imitates `GLSL's swizzling`_, a way to reorder vector values:

        .. code-block:: python

           >>> from arcade.types import Color
           >>> color = Color(180, 90, 0, 255)
           >>> color.swizzle("abgr")
           (255, 0, 90, 180)

        Unlike GLSL, this function allows upper case.

        .. code-block:: python

           >>> from arcade.types import Color
           >>> color = Color(180, 90, 0, 255)
           # You can repeat channels and use upper case
           >>> color.swizzle("ABGRa")
           (255, 0, 90, 180, 255)

        .. note:: The ``order`` is case-insensitive.

                  If you were hoping ``order`` would also specify how to
                  convert data, you may instead be looking for Python's
                  built-in :py:mod:`struct` and :py:mod:`array` modules.

        Args:
            order:
                A string of channel names as letters in ``"RGBArgba"`` with repeats allowed.
        Returns:
            A tuple of channel values in the given ``order``.
        """
        ret = []
        for c in order.lower():
            if c not in "rgba":
                raise ValueError(
                    f"Swizzle string must only contain characters in [RGBArgba], not {c}."
                )
            ret.append(getattr(self, c))
        return tuple(ret)



from __future__ import annotations

from collections.abc import Iterable
from typing import TypeVar

from pyglet.event import EVENT_UNHANDLED

import arcade
from arcade import XYWH
from arcade.gui.events import (
    UIEvent,
    UIMouseDragEvent,
    UIMouseEvent,
    UIMouseMovementEvent,
    UIMousePressEvent,
    UIMouseReleaseEvent,
    UIMouseScrollEvent,
)
from arcade.gui.property import Property, bind
from arcade.gui.surface import Surface
from arcade.gui.widgets import UIWidget
from arcade.gui.widgets.layout import UILayout
from arcade.types import LBWH

W = TypeVar("W", bound="UIWidget")


class UIScrollBar(UIWidget):
    """Scroll bar for a UIScrollLayout.

    Indicating the current view position of the scroll area.
    Supports mouse dragging to scroll the content.
    """

    _thumb_hover = Property(False)
    _dragging = Property(False)

    def __init__(self, scroll_area: UIScrollArea, vertical: bool = True):
        size_hint = (0.05, 1) if vertical else (1, 0.05)

        super().__init__(size_hint=size_hint)
        self.scroll_area = scroll_area
        self.with_background(color=arcade.color.LIGHT_GRAY)
        self.with_border(color=arcade.uicolor.GRAY_CONCRETE)
        self.vertical = vertical

        bind(self, "_thumb_hover", self.trigger_render)
        bind(self, "_dragging", self.trigger_render)
        bind(scroll_area, "scroll_x", self.trigger_full_render)
        bind(scroll_area, "scroll_y", self.trigger_full_render)
        bind(scroll_area, "content_height", self.trigger_full_render)
        bind(scroll_area, "content_width", self.trigger_full_render)

    def on_event(self, event: UIEvent) -> bool | None:
        # check if we are scrollable
        if not self._scrollable():
            return EVENT_UNHANDLED

        # detect if event is mouse down and inside the scroll thumb
        # if so, start dragging the thumb
        thumb_rect_relative = self._thumb_rect()
        thumb_rect = thumb_rect_relative.move(*self.rect.bottom_left)

        if isinstance(event, UIMouseMovementEvent):
            self._thumb_hover = thumb_rect.point_in_rect(event.pos)

        if isinstance(event, UIMousePressEvent) and thumb_rect.point_in_rect(event.pos):
            self._dragging = True
            return True

        # detect if event is mouse drag and thumb is being dragged
        # if so, update the scroll position
        if isinstance(event, UIMouseDragEvent) and self._dragging:
            sx, sy = event.pos - self.rect.bottom_left
            sx -= self._scroll_bar_size() / 2
            sy -= self._scroll_bar_size() / 2

            scroll_area = self.scroll_area

            if self.vertical:
                available_track_size = self.content_height - self._scroll_bar_size()
                target_progress = 1 - sy / available_track_size
                target_progress = max(0, min(1, target_progress))

                scroll_range = scroll_area.surface.height - scroll_area.content_height
                scroll_area.scroll_y = -target_progress * scroll_range
            else:
                available_track_size = self.content_width - self._scroll_bar_size()
                target_progress = sx / available_track_size
                target_progress = max(0, min(1, target_progress))
                scroll_range = scroll_area.surface.width - scroll_area.content_width
                scroll_area.scroll_x = -target_progress * scroll_range

            return True

        # detect if event is mouse up and thumb is being dragged
        # if so, stop dragging the thumb
        if isinstance(event, UIMouseReleaseEvent) and self._dragging:
            self._dragging = False
            return True

        return EVENT_UNHANDLED

    def _scroll_bar_size(self):
        # based on: https://stackoverflow.com/a/16367035

        content_size = (
            self.scroll_area.surface.height if self.vertical else self.scroll_area.surface.width
        )
        view_size = (
            self.scroll_area.content_height if self.vertical else self.scroll_area.content_width
        )
        ratio = view_size / content_size

        scoll_range = self.content_height if self.vertical else self.content_width

        return scoll_range * ratio

    def _scrollable(self):
        return (
            self.scroll_area.surface.height - self.scroll_area.content_height
            if self.vertical
            else self.scroll_area.surface.width - self.scroll_area.content_width
        ) > 0

    def _thumb_rect(self):
        """Calculate the rect of the thumb."""
        scroll_area = self.scroll_area

        scroll_value = scroll_area.scroll_y if self.vertical else scroll_area.scroll_x
        scroll_range = (
            scroll_area.surface.height - scroll_area.content_height
            if self.vertical
            else scroll_area.surface.width - scroll_area.content_width
        )

        if not self._scrollable():
            # content is smaller than the scroll area, full size thumb
            return LBWH(0, 0, self.content_width, self.content_height)

        scroll_progress = -scroll_value / scroll_range

        content_size = self.content_height if self.vertical else self.content_width
        available_track_size = content_size - self._scroll_bar_size()

        if self.vertical:
            scroll_bar_y = self._scroll_bar_size() / 2 + available_track_size * (
                1 - scroll_progress
            )
            scroll_bar_x = self.content_width / 2
            return XYWH(scroll_bar_x, scroll_bar_y, self.content_width, self._scroll_bar_size())

        else:
            scroll_bar_x = self._scroll_bar_size() / 2 + available_track_size * scroll_progress
            scroll_bar_y = self.content_height / 2
            return XYWH(scroll_bar_x, scroll_bar_y, self._scroll_bar_size(), self.content_height)

    def do_render(self, surface: Surface):
        """Render the scroll bar."""
        self.prepare_render(surface)

        if self._dragging:
            thumb_color = arcade.uicolor.DARK_BLUE_WET_ASPHALT
        elif self._thumb_hover:
            thumb_color = arcade.uicolor.GRAY_CONCRETE
        else:
            thumb_color = arcade.uicolor.GRAY_ASBESTOS

        # draw the thumb
        arcade.draw_rect_filled(
            rect=self._thumb_rect(),
            color=thumb_color,
        )


class UIScrollArea(UILayout):
    """A widget that can scroll its children.

    This widget is highly experimental and only provides a proof of concept.

    Args:
        x: x position of the widget
        y: y position of the widget
        width: width of the widget
        height: height of the widget
        children: children of the widget
        size_hint: size hint of the widget
        size_hint_min: minimum size hint of the widget
        size_hint_max: maximum size hint of the widget
        canvas_size: size of the canvas, which is scrollable
        overscroll_x: allow over scrolling in x direction (scroll past the end)
        overscroll_y: allow over scrolling in y direction (scroll past the end)
        **kwargs: passed to UIWidget
    """

    scroll_x = Property[float](default=0.0)
    scroll_y = Property[float](default=0.0)

    scroll_speed = 1.8
    invert_scroll = False

    def __init__(
        self,
        *,
        x: float = 0,
        y: float = 0,
        width: float = 300,
        height: float = 300,
        children: Iterable[UIWidget] = tuple(),
        size_hint=None,
        size_hint_min=None,
        size_hint_max=None,
        canvas_size=(300, 300),
        overscroll_x=False,
        overscroll_y=False,
        **kwargs,
    ):
        super().__init__(
            x=x,
            y=y,
            width=width,
            height=height,
            children=children,
            size_hint=size_hint,
            size_hint_min=size_hint_min,
            size_hint_max=size_hint_max,
            **kwargs,
        )
        self.default_anchor_x = "left"
        self.default_anchor_y = "bottom"
        self.overscroll_x = overscroll_x
        self.overscroll_y = overscroll_y

        self.surface = Surface(
            size=canvas_size,
        )

        bind(self, "scroll_x", self.trigger_full_render)
        bind(self, "scroll_y", self.trigger_full_render)

    def add(self, child: W, **kwargs) -> W:
        """Add a child to the widget."""
        if self._children:
            raise ValueError("UIScrollArea can only have one child")

        super().add(child, **kwargs)
        self.trigger_full_render()

        return child

    def remove(self, child: UIWidget):
        """Remove a child from the widget."""
        super().remove(child)
        self.trigger_full_render()

    def do_layout(self):
        """Layout the children of the widget."""
        total_min_x = 0
        total_min_y = 0

        for child in self.children:
            new_rect = child.rect
            # apply sizehint
            shw, shh = child.size_hint or (None, None)
            # default min_size to be at least 1 for w and h, required by surface
            shw_min, shh_min = child.size_hint_min or (1, 1)
            shw_max, shh_max = child.size_hint_max or (None, None)

            if shw is not None:
                new_width = shw * self.content_width

                new_width = max(shw_min or 1, new_width)
                if shw_max is not None:
                    new_width = min(shw_max, new_width)
                new_rect = new_rect.resize(width=new_width)

            if shh is not None:
                new_height = shh * self.content_height
                new_height = max(shh_min or 1, new_height)

                if shh_max is not None:
                    new_height = min(shh_max, new_height)
                new_rect = new_rect.resize(height=new_height)

            new_rect = new_rect.align_top(self.surface.height).align_left(0)
            total_min_x = max(total_min_x, new_rect.width)
            total_min_y = max(total_min_y, new_rect.height)

            if new_rect != child.rect:
                child.rect = new_rect

        total_min_x = round(total_min_x)
        total_min_y = round(total_min_y)

        # resize surface to fit all children
        if self.surface.size != (total_min_x, total_min_y):
            self.surface.resize(
                size=(total_min_x, total_min_y), pixel_ratio=self.surface.pixel_ratio
            )
            self.scroll_x = 0
            self.scroll_y = 0

    def _do_render(self, surface: Surface, force=False) -> bool:
        if not self.visible:
            return False

        should_render = force or self._requires_render
        rendered = False

        with self.surface.activate():
            if should_render:
                self.surface.clear()

            if self.visible:
                for child in self.children:
                    rendered |= child._do_render(self.surface, should_render)

        if should_render or rendered:
            rendered = True
            self.do_render_base(surface)
            self.do_render(surface)
            self._rendered = True
            self._requires_render = False

        return rendered

    def do_render(self, surface: Surface):
        """Renders the scolled surface into the given surface."""
        self.prepare_render(surface)

        offset_x, offset_y = self._get_scroll_offset()
        # position surface and draw visible area
        self.surface.position = offset_x, offset_y
        self.surface.draw(LBWH(-offset_x, -offset_y, self.content_width, self.content_height))

    def _get_scroll_offset(self):
        """calculates the scroll offset for the surface position,
        also used for calculating mouse event offset."""
        normal_pos_y = self.surface.height - self.content_height

        return self.scroll_x, -normal_pos_y - self.scroll_y

    def on_event(self, event: UIEvent) -> bool | None:
        """Handle scrolling of the widget."""
        if isinstance(event, UIMouseDragEvent) and not self.rect.point_in_rect(event.pos):
            return EVENT_UNHANDLED

        if isinstance(event, UIMouseScrollEvent) and self.rect.point_in_rect(event.pos):
            invert = -1 if self.invert_scroll else 1

            self.scroll_x -= -event.scroll_x * self.scroll_speed * invert
            self.scroll_y -= event.scroll_y * self.scroll_speed * invert

            # clip scrolling to canvas size
            if not self.overscroll_x:
                # clip scroll_x between 0 and -(self.surface.width - self.width)
                scroll_range = int(self.content_width - self.surface.width)
                scroll_range = min(0, scroll_range)  # clip to 0 if content is smaller than surface
                self.scroll_x = min(0, self.scroll_x)
                self.scroll_x = max(self.scroll_x, scroll_range)

            if not self.overscroll_y:
                # clip scroll_y between 0 and -(self.surface.height - self.height)
                scroll_range = int(self.content_height - self.surface.height)
                scroll_range = min(0, scroll_range)  # clip to 0 if content is smaller than surface
                self.scroll_y = min(0, self.scroll_y)
                self.scroll_y = max(self.scroll_y, scroll_range)

            return True

        child_event = event
        if isinstance(event, UIMouseEvent):
            if self.rect.point_in_rect(event.pos):
                # create a new event with the position relative to the child
                off_x, off_y = self._get_scroll_offset()

                child_event = type(event)(**event.__dict__)  # type: ignore
                child_event.x = int(event.x - self.left - off_x)
                child_event.y = int(event.y - self.bottom - off_y)

            else:
                # event is outside the scroll area, do not pass it to the children
                return EVENT_UNHANDLED

        return super().on_event(child_event)



"""
This simple animation example shows how to bounce a rectangle
on the screen.

If Python and Arcade are installed, this example can be run
from the command line with:
python -m arcade.examples.bouncing_rectangle
"""

import arcade

# --- Set up the constants

# Size of the screen
WINDOW_WIDTH = 600
WINDOW_HEIGHT = 600
WINDOW_TITLE = "Bouncing Rectangle Example"

# Rectangle info
RECT_WIDTH = 50
RECT_HEIGHT = 50
RECT_COLOR = arcade.color.DARK_BROWN

BACKGROUND_COLOR = arcade.color.ALMOND


class Item:
    """ This class represents our rectangle """

    def __init__(self):

        # Set up attribute variables

        # Where we are
        self.center_x = 0
        self.center_y = 0

        # Where we are going
        self.change_x = 0
        self.change_y = 0

    def update(self):
        # Move the rectangle
        self.center_x += self.change_x
        self.center_y += self.change_y
        # Check if we need to bounce of right edge
        if self.center_x > WINDOW_WIDTH - RECT_WIDTH / 2:
            self.change_x *= -1
        # Check if we need to bounce of top edge
        if self.center_y > WINDOW_HEIGHT - RECT_HEIGHT / 2:
            self.change_y *= -1
        # Check if we need to bounce of left edge
        if self.center_x < RECT_WIDTH / 2:
            self.change_x *= -1
        # Check if we need to bounce of bottom edge
        if self.center_y < RECT_HEIGHT / 2:
            self.change_y *= -1

    def draw(self):
        # Draw the rectangle
        arcade.draw_rect_filled(
            arcade.rect.XYWH(self.center_x, self.center_y, RECT_WIDTH, RECT_HEIGHT), RECT_COLOR)


class GameView(arcade.View):
    """ Main application class. """

    def __init__(self):
        super().__init__()

        # Create our rectangle
        self.item = Item()
        self.item.center_x = 200
        self.item.center_y = 300
        self.item.change_x = 2
        self.item.change_y = 3

        # Set background color
        self.background_color = BACKGROUND_COLOR

    def on_update(self, delta_time):
        # Move the rectangle
        self.item.update()

    def on_draw(self):
        """ Render the screen. """

        # Clear screen
        self.clear()
        # Draw the rectangle
        self.item.draw()


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create the GameView
    game = GameView()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



"""
Drawing commands that use vertex buffer objects (VBOs).

This module contains commands for basic graphics drawing commands,
but uses Vertex Buffer Objects. This keeps the vertices loaded on
the graphics card for much faster render times.
"""

import itertools
import math
from array import array
from collections import OrderedDict
from collections.abc import Iterable, Sequence
from typing import (
    Generic,
    TypeVar,
    cast,
)

import pyglet.gl as gl

from arcade import ArcadeContext, get_points_for_thick_line, get_window
from arcade.gl import Buffer, BufferDescription, Geometry, Program
from arcade.math import rotate_point
from arcade.types import RGBA255, Color, Point, PointList
from arcade.utils import copy_dunders_unimplemented

__all__ = [
    "Shape",
    "create_line",
    "create_line_generic_with_colors",
    "create_line_generic",
    "create_line_strip",
    "create_line_loop",
    "create_lines",
    "create_lines_with_colors",
    "create_polygon",
    "create_rectangle_filled",
    "create_rectangle_outline",
    "get_rectangle_points",
    "create_rectangle",
    "create_rectangle_filled_with_colors",
    "create_rectangles_filled_with_colors",
    "create_triangles_filled_with_colors",
    "create_triangles_strip_filled_with_colors",
    "create_ellipse_filled",
    "create_ellipse_outline",
    "create_ellipse",
    "create_ellipse_filled_with_colors",
    "ShapeElementList",
]


@copy_dunders_unimplemented  # Temp fix for https://github.com/pythonarcade/arcade/issues/2074
class Shape:
    """
    A container for arbitrary geometry representing a shape.

    This shape can be drawn using the draw() method, or added to a
    ShapeElementList for drawing in batch.

    Args:
        points: A list of points that make up the shape.
        colors: A list of colors that correspond to the points.
        mode: The OpenGL drawing mode. Defaults to GL_TRIANGLES.
        program: The program to use when drawing this shape (Shape.draw() only)
    """

    def __init__(
        self,
        points: PointList,
        colors: Sequence[RGBA255],
        # vao: Geometry,
        # vbo: Buffer,
        mode: int = gl.GL_TRIANGLES,
        program: Program | None = None,
    ) -> None:
        self.ctx = get_window().ctx
        self.program = program or self.ctx.line_generic_with_colors_program
        self.mode = mode

        if len(points) != len(colors):
            raise ValueError("Number of points and colors must match.")

        self.points = points
        # Ensure colors have 4 components
        self.colors = [Color.from_iterable(color) for color in colors]
        # Pack the data into a single array
        self.data = array("f", [c for a in zip(self.points, self.colors) for b in a for c in b])
        self.vertices = len(points)

        self.geometry: Geometry | None = None
        self.buffer: Buffer | None = None

    def _init_geometry(self) -> None:
        # NOTE: When drawing a single shape we're not using an index buffer
        self.buffer = self.program.ctx.buffer(data=self.data)
        self.geometry = self.ctx.geometry(
            [
                BufferDescription(
                    self.buffer,
                    "2f 4f",
                    ("in_vert", "in_color"),
                ),
            ]
        )

    def draw(self) -> None:
        """
        Draw this shape. Drawing this way isn't as fast as drawing multiple
        shapes batched together in a ShapeElementList.
        """
        if self.geometry is None:
            self._init_geometry()

        if self.geometry is not None:
            self.geometry.render(self.program, mode=self.mode)


def create_line(
    start_x: float,
    start_y: float,
    end_x: float,
    end_y: float,
    color: RGBA255,
    line_width: float = 1,
) -> Shape:
    """
    Create a Shape object for a line.

    Args:
        start_x: Starting x position
        start_y: Starting y position
        end_x: Ending x position
        end_y: Ending y position
        color: Color of the line
        line_width: Width of the line
    """
    points = get_points_for_thick_line(start_x, start_y, end_x, end_y, line_width)
    color_list = [color, color, color, color]
    triangle_point_list = points[1], points[0], points[2], points[3]
    return create_triangles_strip_filled_with_colors(triangle_point_list, color_list)


def create_line_generic_with_colors(
    point_list: PointList,
    color_sequence: Sequence[RGBA255],
    shape_mode: int,
) -> Shape:
    """
    This function is used by ``create_line_strip`` and ``create_line_loop``,
    just changing the OpenGL type for the line drawing.

    Args:
        point_list: A list of points that make up the shape.
        color_sequence: A sequence of colors such
            as a :py:class:`list`; each color must be either a
            :py:class:`~arcade.types.Color` instance or a 4-length RGBA
            :py:class:`tuple`.
        shape_mode: The OpenGL drawing mode. Defaults to ``GL_TRIANGLES``.
    """
    return Shape(
        points=point_list,
        colors=color_sequence,
        mode=shape_mode,
    )


def create_line_generic(
    point_list: PointList,
    color: RGBA255,
    shape_mode: int,
) -> Shape:
    """
    This function is used by ``create_line_strip`` and ``create_line_loop``,
    just changing the OpenGL type for the line drawing.

    Args:
        point_list: A list of points that make up the shape.
        color: A color such as a :py:class:`~arcade.types.Color`
        shape_mode: The OpenGL drawing mode. Defaults to ``GL_TRIANGLES``.
    """
    colors = [Color.from_iterable(color)] * len(point_list)
    return create_line_generic_with_colors(point_list, colors, shape_mode)


def create_line_strip(point_list: PointList, color: RGBA255, line_width: float = 1) -> Shape:
    """
    Create a multi-point line to be rendered later. This works faster than draw_line because
    the vertexes are only loaded to the graphics card once, rather than each frame.

    Internally, thick lines are created by two triangles.

    Args:
        point_list: A list of points that make up the shape.
        color: A color such as a :py:class:`~arcade.types.Color`
        line_width: Width of the line
    """
    if line_width == 1:
        return create_line_generic(point_list, color, gl.GL_LINE_STRIP)

    triangle_point_list: list[Point] = []
    new_color_list: list[RGBA255] = []
    for i in range(1, len(point_list)):
        start_x = point_list[i - 1][0]
        start_y = point_list[i - 1][1]
        end_x = point_list[i][0]
        end_y = point_list[i][1]
        color1 = color
        color2 = color
        points = get_points_for_thick_line(start_x, start_y, end_x, end_y, line_width)
        new_color_list += color1, color2, color1, color2
        triangle_point_list += points[1], points[0], points[2], points[3]

    return create_triangles_strip_filled_with_colors(triangle_point_list, new_color_list)


def create_line_loop(
    point_list: PointList,
    color: RGBA255,
    line_width: float = 1,
) -> Shape:
    """
    Create a multi-point line loop to be rendered later. This works faster than draw_line because
    the vertexes are only loaded to the graphics card once, rather than each frame.

    Args:
        point_list: A list of points that make up the shape.
        color: A color such as a :py:class:`~arcade.types.Color`
        line_width: Width of the line
    """
    point_list = list(point_list) + [point_list[0]]
    return create_line_strip(point_list, color, line_width)


def create_lines(
    point_list: PointList,
    color: RGBA255,
) -> Shape:
    """
    Create a multi-point line loop to be rendered later. This works faster than draw_line because
    the vertexes are only loaded to the graphics card once, rather than each frame.

    Args:
        point_list: A list of points that make up the shape.
        color: A color such as a :py:class:`~arcade.types.Color`
    """
    return create_line_generic(point_list, color, gl.GL_LINES)


def create_lines_with_colors(
    point_list: PointList,
    color_list: Sequence[RGBA255],
    line_width: float = 1,
) -> Shape:
    """
    Create a line segments to be rendered later. This works faster than draw_line because
    the vertexes are only loaded to the graphics card once, rather than each frame.

    Args:
        point_list: Line segments start and end point tuples list
        color_list: Three or four byte tuples list for every point
        line_width: Width of the line
    """
    if line_width == 1:
        return create_line_generic_with_colors(point_list, color_list, gl.GL_LINES)

    triangle_point_list: list[Point] = []
    new_color_list: list[RGBA255] = []
    for i in range(1, len(point_list), 2):
        start_x = point_list[i - 1][0]
        start_y = point_list[i - 1][1]
        end_x = point_list[i][0]
        end_y = point_list[i][1]
        color1 = color_list[i - 1]
        color2 = color_list[i]
        points = get_points_for_thick_line(start_x, start_y, end_x, end_y, line_width)
        new_color_list += color1, color1, color2, color1, color2, color2
        triangle_point_list += points[0], points[1], points[2], points[0], points[2], points[3]

    return create_triangles_filled_with_colors(triangle_point_list, new_color_list)


def create_polygon(point_list: PointList, color: RGBA255) -> Shape:
    """
    Draw a convex polygon. This will NOT draw a concave polygon.
    Because of this, you might not want to use this function.

    The function returns a Shape object that can be drawn with ``my_shape.draw()``.
    Don't create the shape in the draw method, create it in the setup method and then
    draw it in ``on_draw``.

    For even faster performance, add multiple shapes into a ShapeElementList and
    draw that list. This allows nearly unlimited shapes to be drawn just as fast
    as one.

    Args:
        point_list: A list of points that make up the shape.
        color: A color such as a :py:class:`~arcade.types.Color`
    """
    # We assume points were given in order, either clockwise or counter clockwise.
    # Polygon is assumed to be monotone.
    # To fill the polygon, we start by one vertex, and we chain triangle strips
    # alternating with vertices to the left and vertices to the right of the
    # initial vertex.
    half = len(point_list) // 2
    interleaved = itertools.chain.from_iterable(
        itertools.zip_longest(point_list[:half], reversed(point_list[half:]))
    )
    point_list = [p for p in interleaved if p is not None]
    return create_line_generic(point_list, color, gl.GL_TRIANGLE_STRIP)


def create_rectangle_filled(
    center_x: float,
    center_y: float,
    width: float,
    height: float,
    color: RGBA255,
    tilt_angle: float = 0,
) -> Shape:
    """
    Create a filled rectangle.

    The function returns a Shape object that can be drawn with ``my_shape.draw()``.
    Don't create the shape in the draw method, create it in the setup method and then
    draw it in ``on_draw``.

    For even faster performance, add multiple shapes into a ShapeElementList and
    draw that list. This allows nearly unlimited shapes to be drawn just as fast
    as one.

    Args:
        center_x: X position of the center of the rectangle
        center_y: Y position of the center of the rectangle
        width: Width of the rectangle
        height: Height of the rectangle
        color: A color such as a :py:class:`~arcade.types.Color`
        tilt_angle: Angle to tilt the rectangle in degrees
    """
    return create_rectangle(
        center_x,
        center_y,
        width,
        height,
        color,
        tilt_angle=tilt_angle,
    )


def create_rectangle_outline(
    center_x: float,
    center_y: float,
    width: float,
    height: float,
    color: RGBA255,
    border_width: float = 1,
    tilt_angle: float = 0,
) -> Shape:
    """
    Create a rectangle outline.

    The function returns a Shape object that can be drawn with ``my_shape.draw()``.
    Don't create the shape in the draw method, create it in the setup method and then
    draw it in ``on_draw``.

    For even faster performance, add multiple shapes into a ShapeElementList and
    draw that list. This allows nearly unlimited shapes to be drawn just as fast
    as one.

    Args:
        center_x: X position of the center of the rectangle
        center_y: Y position of the center of the rectangle
        width: Width of the rectangle
        height: Height of the rectangle
        color: A color such as a :py:class:`~arcade.types.Color`
        border_width: Width of the border
        tilt_angle: Angle to tilt the rectangle in degrees
    """
    return create_rectangle(
        center_x,
        center_y,
        width,
        height,
        color,
        border_width,
        tilt_angle,
        filled=False,
    )


def get_rectangle_points(
    center_x: float,
    center_y: float,
    width: float,
    height: float,
    tilt_angle: float = 0,
) -> PointList:
    """
    Utility function that will return all four coordinate points of a
    rectangle given the x, y center, width, height, and rotation.

    Args:
        center_x: X position of the center of the rectangle
        center_y: Y position of the center of the rectangle
        width: Width of the rectangle
        height: Height of the rectangle
        tilt_angle: Angle to tilt the rectangle in degrees
    """
    x1 = -width / 2 + center_x
    y1 = -height / 2 + center_y

    x2 = -width / 2 + center_x
    y2 = height / 2 + center_y

    x3 = width / 2 + center_x
    y3 = height / 2 + center_y

    x4 = width / 2 + center_x
    y4 = -height / 2 + center_y

    if tilt_angle:
        x1, y1 = rotate_point(x1, y1, center_x, center_y, tilt_angle)
        x2, y2 = rotate_point(x2, y2, center_x, center_y, tilt_angle)
        x3, y3 = rotate_point(x3, y3, center_x, center_y, tilt_angle)
        x4, y4 = rotate_point(x4, y4, center_x, center_y, tilt_angle)

    return [(x1, y1), (x2, y2), (x3, y3), (x4, y4)]


def create_rectangle(
    center_x: float,
    center_y: float,
    width: float,
    height: float,
    color: RGBA255,
    border_width: float = 1,
    tilt_angle: float = 0,
    filled=True,
) -> Shape:
    """
    This function creates a rectangle using a vertex buffer object.

    The function returns a Shape object that can be drawn with ``my_shape.draw()``.
    Don't create the shape in the draw method, create it in the setup method and then
    draw it in ``on_draw``.

    For even faster performance, add multiple shapes into a ShapeElementList and
    draw that list. This allows nearly unlimited shapes to be drawn just as fast
    as one.

    Args:
        center_x: X position of the center of the rectangle
        center_y: Y position of the center of the rectangle
        width: Width of the rectangle
        height: Height of the rectangle
        color: A color such as a :py:class:`~arcade.types.Color`
        border_width: Width of the border
        tilt_angle: Angle to tilt the rectangle in degrees
        filled: If True, the rectangle is filled. If False, it is an outline.
    """
    data: list[Point] = cast(
        list[Point], get_rectangle_points(center_x, center_y, width, height, tilt_angle)
    )

    if filled:
        data[-2:] = reversed(data[-2:])
    else:
        i_lb = (
            center_x - width / 2 + border_width / 2,
            center_y - height / 2 + border_width / 2,
        )
        i_rb = (
            center_x + width / 2 - border_width / 2,
            center_y - height / 2 + border_width / 2,
        )
        i_rt = (
            center_x + width / 2 - border_width / 2,
            center_y + height / 2 - border_width / 2,
        )
        i_lt = (
            center_x - width / 2 + border_width / 2,
            center_y + height / 2 - border_width / 2,
        )

        o_lb = (
            center_x - width / 2 - border_width / 2,
            center_y - height / 2 - border_width / 2,
        )
        o_rb = (
            center_x + width / 2 + border_width / 2,
            center_y - height / 2 - border_width / 2,
        )
        o_rt = (
            center_x + width / 2 + border_width / 2,
            center_y + height / 2 + border_width / 2,
        )
        o_lt = (
            center_x - width / 2 - border_width / 2,
            center_y + height / 2 + border_width / 2,
        )

        data = [o_lt, i_lt, o_rt, i_rt, o_rb, i_rb, o_lb, i_lb, o_lt, i_lt]

        if tilt_angle != 0:
            point_list_2: list[Point] = []
            for point in data:
                new_point = rotate_point(point[0], point[1], center_x, center_y, tilt_angle)
                point_list_2.append(new_point)
            data = point_list_2

        border_width = 1

    shape_mode = gl.GL_TRIANGLE_STRIP
    return create_line_generic(data, color, shape_mode)


def create_rectangle_filled_with_colors(point_list, color_list) -> Shape:
    """
    This function creates one rectangle/quad using a vertex buffer object.

    The function returns a Shape object that can be drawn with ``my_shape.draw()``.
    Don't create the shape in the draw method, create it in the setup method and then
    draw it in ``on_draw``.

    For even faster performance, add multiple shapes into a ShapeElementList and
    draw that list. This allows nearly unlimited shapes to be drawn just as fast
    as one.

    Args:
        point_list: List of points to create the rectangle from
        color_list: List of colors to create the rectangle from
    """
    shape_mode = gl.GL_TRIANGLE_STRIP
    new_point_list = [point_list[0], point_list[1], point_list[3], point_list[2]]
    new_color_list = [color_list[0], color_list[1], color_list[3], color_list[2]]
    return create_line_generic_with_colors(new_point_list, new_color_list, shape_mode)


def create_rectangles_filled_with_colors(point_list, color_list: Sequence[RGBA255]) -> Shape:
    """
    This function creates multiple rectangle/quads using a vertex buffer object.

    The function returns a Shape object that can be drawn with ``my_shape.draw()``.
    Don't create the shape in the draw method, create it in the setup method and then
    draw it in ``on_draw``.

    For even faster performance, add multiple shapes into a ShapeElementList and
    draw that list. This allows nearly unlimited shapes to be drawn just as fast
    as one.

    Args:
        point_list: List of points to create the rectangles from
        color_list: List of colors to create the rectangles from
    """
    shape_mode = gl.GL_TRIANGLES
    new_point_list: list[Point] = []
    new_color_list: list[RGBA255] = []
    for i in range(0, len(point_list), 4):
        new_point_list += [point_list[0 + i], point_list[1 + i], point_list[3 + i]]
        new_point_list += [point_list[1 + i], point_list[3 + i], point_list[2 + i]]

        new_color_list += [color_list[0 + i], color_list[1 + i], color_list[3 + i]]
        new_color_list += [color_list[1 + i], color_list[3 + i], color_list[2 + i]]

    return create_line_generic_with_colors(new_point_list, new_color_list, shape_mode)


def create_triangles_filled_with_colors(
    point_list: PointList,
    color_sequence: Sequence[RGBA255],
) -> Shape:
    """
    This function creates multiple triangles using a vertex buffer object.
    Triangles are build for every 3 sequential vertices with step of 3 vertex
    Total amount of triangles to be rendered: len(point_list) / 3

    The function returns a Shape object that can be drawn with ``my_shape.draw()``.
    Don't create the shape in the draw method, create it in the setup method and then
    draw it in ``on_draw``.

    For even faster performance, add multiple shapes into a ShapeElementList and
    draw that list. This allows nearly unlimited shapes to be drawn just as fast
    as one.

    Args:
        point_list: Triangles vertices tuples.
        color_sequence: A sequence of colors such
            as a :py:class:`list`; each color must be either a
            :py:class:`~arcade.types.Color` instance or a 4-length RGBA
            :py:class:`tuple`.
    """
    shape_mode = gl.GL_TRIANGLES
    return create_line_generic_with_colors(point_list, color_sequence, shape_mode)


def create_triangles_strip_filled_with_colors(
    point_list,
    color_sequence: Sequence[RGBA255],
) -> Shape:
    """
    This function creates multiple triangles using a vertex buffer object.
    Triangles are built for every 3 sequential vertices with step of 1 vertex
    Total amount of triangles to be rendered: len(point_list) - 2

    The function returns a Shape object that can be drawn with ``my_shape.draw()``.
    Don't create the shape in the draw method, create it in the setup method and then
    draw it in ``on_draw``.

    For even faster performance, add multiple shapes into a ShapeElementList and
    draw that list. This allows nearly unlimited shapes to be drawn just as fast
    as one.

    Args:
        point_list: Triangles vertices tuples.
        color_sequence: A sequence of colors such
            as a :py:class:`list`; each color must be either a
            :py:class:`~arcade.types.Color` instance or a 4-length RGBA
            :py:class:`tuple`.
    """
    shape_mode = gl.GL_TRIANGLE_STRIP
    return create_line_generic_with_colors(point_list, color_sequence, shape_mode)


def create_ellipse_filled(
    center_x: float,
    center_y: float,
    width: float,
    height: float,
    color: RGBA255,
    tilt_angle: float = 0,
    num_segments: int = 128,
) -> Shape:
    """
    Create a filled ellipse. Or circle if you use the same width and height.

    The function returns a Shape object that can be drawn with ``my_shape.draw()``.
    Don't create the shape in the draw method, create it in the setup method and then
    draw it in ``on_draw``.

    For even faster performance, add multiple shapes into a ShapeElementList and
    draw that list. This allows nearly unlimited shapes to be drawn just as fast
    as one.

    Args:
        center_x: X position of the center of the ellipse
        center_y: Y position of the center of the ellipse
        width: Width of the ellipse
        height: Height of the ellipse
        color: A color such as a :py:class:`~arcade.types.Color`
        tilt_angle: Angle to tilt the ellipse
        num_segments: Number of segments to use to draw the ellipse
    """
    border_width = 1
    return create_ellipse(
        center_x,
        center_y,
        width,
        height,
        color,
        border_width,
        tilt_angle,
        num_segments,
        filled=True,
    )


def create_ellipse_outline(
    center_x: float,
    center_y: float,
    width: float,
    height: float,
    color: RGBA255,
    border_width: float = 1,
    tilt_angle: float = 0,
    num_segments: int = 128,
) -> Shape:
    """
    Create an outline of an ellipse.

    The function returns a Shape object that can be drawn with ``my_shape.draw()``.
    Don't create the shape in the draw method, create it in the setup method and then
    draw it in ``on_draw``.

    For even faster performance, add multiple shapes into a ShapeElementList and
    draw that list. This allows nearly unlimited shapes to be drawn just as fast
    as one.

    Args:
        center_x: X position of the center of the ellipse
        center_y: Y position of the center of the ellipse
        width: Width of the ellipse
        height: Height of the ellipse
        color: A color such as a :py:class:`~arcade.types.Color`
        border_width: Width of the border
        tilt_angle: Angle to tilt the ellipse
        num_segments: Number of segments to use to draw the ellipse
    """
    return create_ellipse(
        center_x,
        center_y,
        width,
        height,
        color,
        border_width,
        tilt_angle,
        num_segments,
        filled=False,
    )


def create_ellipse(
    center_x: float,
    center_y: float,
    width: float,
    height: float,
    color: RGBA255,
    border_width: float = 1,
    tilt_angle: float = 0,
    num_segments: int = 32,
    filled: bool = True,
) -> Shape:
    """
    This creates an ellipse vertex buffer object (VBO).

    The function returns a Shape object that can be drawn with ``my_shape.draw()``.
    Don't create the shape in the draw method, create it in the setup method and then
    draw it in ``on_draw``.

    For even faster performance, add multiple shapes into a ShapeElementList and
    draw that list. This allows nearly unlimited shapes to be drawn just as fast
    as one.

    Args:
        center_x: X position of the center of the ellipse.
        center_y: Y position of the center of the ellipse.
        width: Width of the ellipse.
        height: Height of the ellipse.
        color: Color of the ellipse.
        border_width: Width of the border.
        tilt_angle: Angle to tilt the ellipse.
        num_segments: Number of segments to use to draw the ellipse.
        filled: If True, create a filled ellipse. If False, create an outline.
    """
    # Create an array with the vertex point_list
    point_list = []

    for segment in range(num_segments):
        theta = 2.0 * 3.1415926 * segment / num_segments

        x = width / 2 * math.cos(theta) + center_x
        y = height / 2 * math.sin(theta) + center_y

        if tilt_angle:
            x, y = rotate_point(x, y, center_x, center_y, tilt_angle)

        point_list.append((x, y))

    if filled:
        half = len(point_list) // 2
        interleaved = itertools.chain.from_iterable(
            itertools.zip_longest(point_list[:half], reversed(point_list[half:]))
        )
        point_list = [p for p in interleaved if p is not None]
        shape_mode = gl.GL_TRIANGLE_STRIP
    else:
        point_list.append(point_list[0])
        shape_mode = gl.GL_LINE_STRIP

    return create_line_generic(point_list, color, shape_mode)


def create_ellipse_filled_with_colors(
    center_x: float,
    center_y: float,
    width: float,
    height: float,
    outside_color: RGBA255,
    inside_color: RGBA255,
    tilt_angle: float = 0,
    num_segments: int = 32,
) -> Shape:
    """
    Draw an ellipse, and specify inside/outside color. Used for doing gradients.

    The function returns a Shape object that can be drawn with ``my_shape.draw()``.
    Don't create the shape in the draw method, create it in the setup method and then
    draw it in ``on_draw``.

    For even faster performance, add multiple shapes into a ShapeElementList and
    draw that list. This allows nearly unlimited shapes to be drawn just as fast
    as one.

    Args:
        center_x: X position of the center of the ellipse.
        center_y:  Y position of the center of the ellipse.
        width: Width of the ellipse.
        height: Height of the ellipse.
        outside_color: Color of the outside of the ellipse.
        inside_color: Color of the inside of the ellipse.
        tilt_angle: Angle to tilt the ellipse.
        num_segments: Number of segments to use to draw the ellipse.
    """
    # Create an array with the vertex data
    # Create an array with the vertex point_list
    point_list = [(center_x, center_y)]

    for segment in range(num_segments):
        theta = 2.0 * 3.1415926 * segment / num_segments

        x = width * math.cos(theta) + center_x
        y = height * math.sin(theta) + center_y

        if tilt_angle:
            x, y = rotate_point(x, y, center_x, center_y, tilt_angle)

        point_list.append((x, y))
    point_list.append(point_list[1])

    color_list = [inside_color] + [outside_color] * (num_segments + 1)
    return create_line_generic_with_colors(point_list, color_list, gl.GL_TRIANGLE_FAN)


TShape = TypeVar("TShape", bound=Shape)
"""
Type variable for Shape or subclasses.
"""


@copy_dunders_unimplemented
class ShapeElementList(Generic[TShape]):
    """
    A ShapeElementList is a list of shapes that can be drawn together
    in a back for better performance. ShapeElementLists are suited for
    drawing a large number of shapes that are static. If you need to
    move a lot of shapes it's better to use pyglet's shape system.

    Adding new shapes is fast, but removing them is slow.

    Args:
        blend: If True, shapes will be drawn with blending enabled.
    """

    def __init__(self, blend: bool = True) -> None:
        # The context this shape list belongs to
        self.ctx = get_window().ctx
        # List of sprites in the sprite list
        self.shape_list: list[TShape] = []
        self.change_x = 0.0
        self.change_y = 0.0
        self._center_x = 0.0
        self._center_y = 0.0
        self._angle = 0.0
        self.program = self.ctx.shape_element_list_program
        self.batches: dict[int, _Batch] = OrderedDict()
        self.dirties: set[_Batch] = set()

        self._blend = blend

    def append(self, item: TShape) -> None:
        """
        Add a new shape to the list.

        Args:
            item: Shape to add to the list.
        """
        self.shape_list.append(item)
        batch = self.batches.get(item.mode, None)
        if batch is None:
            batch = _Batch(
                self.ctx,
                self.program,
                item.mode,
            )
            self.batches[item.mode] = batch

        batch.append(item)

        # Mark the group as dirty
        self.dirties.add(batch)

    def remove(self, item: TShape) -> None:
        """
        Remove a specific shape from the list.

        Args:
            item: Shape to remove from the list.
        """
        self.shape_list.remove(item)
        batch = self.batches[item.mode]
        batch.remove(item)
        self.dirties.add(batch)

    def update(self) -> None:
        """
        Update the internals of the shape list.
        This is automatically called when you call draw().

        In some instances you may need to call this manually to
        update the shape list before drawing.
        """
        for group in self.dirties:
            group.update()

    def draw(self) -> None:
        """
        Draw all the shapes.
        """
        self.program["Position"] = self._center_x, self._center_y
        self.program["Angle"] = -self._angle

        self.update()
        self.dirties.clear()

        if self._blend:
            self.ctx.enable_only(self.ctx.BLEND)

        # Draw the batches
        for batch in self.batches.values():
            batch.draw()

        if self._blend:
            self.ctx.disable(self.ctx.BLEND)

    def clear(self, position: bool = True, angle: bool = True) -> None:
        """
        Clear all the contents from the shape list.

        Args:
            position: Reset the position to ``0, 0``
            angle: Reset the angle to ``0.0``
        """
        self.shape_list.clear()
        self.batches.clear()
        self.dirties.clear()
        if position:
            self.center_x = 0
            self.center_y = 0
        if angle:
            self.angle = 0

    def move(self, change_x: float, change_y: float) -> None:
        """
        Change the center_x/y of the shape list relative to the current position.

        Args:
            change_x: Amount to move on the x axis
            change_y: Amount to move on the y axis
        """
        self.center_x += change_x
        self.center_y += change_y

    @property
    def position(self) -> tuple[float, float]:
        """
        Get or set the position of the ShapeElementList.

        This is the equivalent of setting center_x and center_y
        """
        return self._center_x, self._center_y

    @position.setter
    def position(self, value: tuple[float, float]) -> None:
        self._center_x, self._center_y = value

    @property
    def center_x(self) -> float:
        """Get or set the center x coordinate of the shape list."""
        return self._center_x

    @center_x.setter
    def center_x(self, value: float) -> None:
        self._center_x = value

    @property
    def center_y(self) -> float:
        """Get or set the center y coordinate of the shape list."""
        return self._center_y

    @center_y.setter
    def center_y(self, value: float) -> None:
        self._center_y = value

    @property
    def angle(self) -> float:
        """Get or set the rotation in degrees (clockwise)"""
        return self._angle

    @angle.setter
    def angle(self, value: float) -> None:
        self._angle = value

    def __len__(self) -> int:
        """Return the length of the shape list."""
        return len(self.shape_list)

    def __iter__(self) -> Iterable[TShape]:
        """Return an iterable object of sprites."""
        return iter(self.shape_list)

    def __getitem__(self, i):
        return self.shape_list[i]


class _Batch(Generic[TShape]):
    """
    A collection of shapes with the same configuration.

    The group uniqueness is based on the primitive mode
    """

    # Flags for keeping track of changes
    ADD = 1
    REMOVE = 3
    # The byte size of a vertex
    VERTEX_SIZE = 4 * 6  # 24 bytes (2 floats for position, 4 floats for color)
    RESET_IDX = 2**32 - 1

    def __init__(
        self,
        ctx: ArcadeContext,
        program: Program,
        mode: int,
    ) -> None:
        self.ctx = ctx
        self.program = program
        self.mode = mode

        self.vbo = self.ctx.buffer(reserve=1024 * self.VERTEX_SIZE, usage="dynamic")
        self.ibo = self.ctx.buffer(reserve=1024 * 4, usage="dynamic")

        self.geometry = self.ctx.geometry(
            content=[
                BufferDescription(
                    self.vbo,
                    "2f 4f",
                    ("in_vert", "in_color"),
                )
            ],
            index_buffer=self.ibo,
        )

        self.items: list[TShape] = []
        self.new_items: list[TShape] = []
        self.vertices = 0  # Total vertices in the batch
        self.elements = 0  # Total elements in the batch
        self.FLAGS = 0  # Flags to indicate changes

    def draw(self) -> None:
        """Draw the batch."""
        if self.elements == 0:
            return

        self.geometry.render(self.program, vertices=self.elements, mode=self.mode)

    def append(self, item: TShape) -> None:
        self.new_items.append(item)
        self.FLAGS |= self.ADD

    def remove(self, item: TShape) -> None:
        self.items.remove(item)
        self.FLAGS |= self.REMOVE

    def update(self) -> None:
        """Update the internals of the batch."""
        if self.FLAGS == 0:
            return

        # If only add flag is set we simply copy in the new data
        if self.FLAGS == self.ADD:
            new_data = array("f")
            new_ibo = array("I")
            counter = itertools.count(self.vertices)
            new_vertices = 0

            # Prepare data for new newly added items
            for item in self.new_items:
                # Update the batch vertex count
                new_vertices += item.vertices
                # Build up an array of new vertex data
                new_data.extend(item.data)

                # Build new index buffer data
                new_ibo.extend(itertools.islice(counter, item.vertices))
                new_ibo.append(self.RESET_IDX)  # Restart the primitive

            # Calculate the size of new and old data
            vbo_old_size = self.vertices * self.VERTEX_SIZE
            vbo_new_data_size = len(new_data) * self.VERTEX_SIZE
            vbo_new_size = vbo_old_size + vbo_new_data_size

            if vbo_new_size > self.vbo.size:
                # Copy out the buffer, resize and copy back
                buff = self.ctx.buffer(reserve=self.vbo.size)
                buff.copy_from_buffer(self.vbo)
                self.vbo.orphan(size=vbo_new_size * 2)
                self.vbo.copy_from_buffer(buff)

            # Calculate the index buffer size
            ibo_old_size = self.elements * 4
            ibo_new_data_size = len(new_ibo) * 4
            ibo_new_size = ibo_old_size + ibo_new_data_size

            if ibo_new_size > self.ibo.size:
                # Copy out the buffer, resize and copy back
                buff = self.ctx.buffer(reserve=self.ibo.size)
                buff.copy_from_buffer(self.ibo)
                self.ibo.orphan(size=ibo_new_size * 2)
                self.ibo.copy_from_buffer(buff)

            # Copy in the new data with offsets
            self.vbo.write(new_data, offset=vbo_old_size)
            self.ibo.write(new_ibo, offset=ibo_old_size)

            self.items.extend(self.new_items)
            self.new_items.clear()
            # Element count is the vertex count + the number of restart indices
            self.vertices += new_vertices
            self.elements = self.vertices + len(self.items)
        else:
            # Do the expensive rebuild
            # NOTE: We don't need to worry about buffer size here
            #       because we know the buffer hasn't grown.
            #       Simply collect the data and write it to the buffer.
            #       and update the vertex count.
            # NOTE: This can be optimized in the future, but pyglet shapes are better.
            self.items.extend(self.new_items)
            self.new_items.clear()

            data = array("f")
            ibo = array("I")
            counter = itertools.count()
            self.vertices = 0
            self.elements = 0

            for item in self.items:
                # Build up an array of new vertex data
                data.extend(item.data)
                # Build new index buffer data
                ibo.extend(itertools.islice(counter, item.vertices))
                ibo.append(self.RESET_IDX)  # Restart the primitive

                self.vertices += item.vertices
                self.elements += item.vertices + 1

            # Resize the buffers if needed
            data_size = self.vertices * self.VERTEX_SIZE
            if data_size > self.vbo.size:
                self.vbo.orphan(size=data_size * 2)

            index_size = self.elements * 4
            if index_size > self.ibo.size:
                self.ibo.orphan(size=index_size * 2)

            self.vbo.write(data)
            self.ibo.write(ibo)

        self.FLAGS = 0



"""
Perspective Parallax

Using a perspective projector and sprites at different depths you can cheaply get a parallax effect.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.perspective_parallax
"""

import math

import arcade

SCREEN_WIDTH, SCREEN_HEIGHT = 800, 600
LAYERS = (
    ":assets:/images/cybercity_background/far-buildings.png",
    ":assets:/images/cybercity_background/back-buildings.png",
    ":assets:/images/cybercity_background/foreground.png",
)


class PerspectiveParallax(arcade.Window):
    def __init__(self):
        super().__init__(SCREEN_WIDTH, SCREEN_HEIGHT, "Perspective Parallax")
        self.t = 0.0
        self.camera = arcade.camera.PerspectiveProjector()

        self.camera_data = self.camera.view
        self.camera_data.zoom = 2.0

        self.camera.projection.far = 1000

        self.background_sprites = arcade.SpriteList()
        for index, layer_src in enumerate(LAYERS):
            layer = arcade.Sprite(layer_src)
            layer.depth = -500 + index * 100.0
            self.background_sprites.append(layer)

    def on_draw(self):
        self.clear()
        with self.camera.activate():
            self.background_sprites.draw(pixelated=True)

    def on_update(self, delta_time: float):
        self.t += delta_time

        self.camera_data.position = (math.cos(self.t) * 200.0, math.sin(self.t) * 200.0, 0.0)


def main():
    window = PerspectiveParallax()
    window.run()


if __name__ == "__main__":
    main()



from arcade.types import RGBOrA255

from .arc import draw_arc_filled, draw_arc_outline


def draw_parabola_filled(
    start_x: float,
    start_y: float,
    end_x: float,
    height: float,
    color: RGBOrA255,
    tilt_angle: float = 0,
) -> None:
    """
    Draws a filled in parabola.

    Args:
        start_x:
            The starting x position of the parabola
        start_y:
            The starting y position of the parabola
        end_x:
            The ending x position of the parabola
        height:
            The height of the parabola
        color:
            A 3 or 4 length tuple of 0-255 channel values
            or a :py:class:`.Color` instance.
        tilt_angle:
            The angle of the tilt of the parabola (clockwise)
    """
    center_x = (start_x + end_x) / 2
    center_y = start_y + height
    start_angle = 0
    end_angle = 180
    width = start_x - end_x
    draw_arc_filled(center_x, center_y, width, height, color, start_angle, end_angle, tilt_angle)


def draw_parabola_outline(
    start_x: float,
    start_y: float,
    end_x: float,
    height: float,
    color: RGBOrA255,
    border_width: float = 1,
    tilt_angle: float = 0,
) -> None:
    """
    Draws the outline of a parabola.

    Args:
        start_x:
            The starting x position of the parabola
        start_y:
            The starting y position of the parabola
        end_x:
            The ending x position of the parabola
        height:
            The height of the parabola
        color:
            A 3 or 4 length tuple of 0-255 channel values
            or a :py:class:`.Color` instance.
        border_width:
            The width of the parabola
        tilt_angle:
            The angle of the tilt of the parabola (clockwise)
    """
    center_x = (start_x + end_x) / 2
    center_y = start_y + height
    start_angle = 0
    end_angle = 180
    width = start_x - end_x
    draw_arc_outline(
        center_x,
        center_y,
        width,
        height,
        color,
        start_angle,
        end_angle,
        border_width,
        tilt_angle,
    )



"""
Use texture layers with shadertoy.
We simply mix the two texture layers.
"""

import arcade
from arcade.experimental.shadertoy import Shadertoy

# Do the math to figure out our screen dimensions
SCREEN_WIDTH = 1280
SCREEN_HEIGHT = 720
SCREEN_TITLE = "ShaderToy Texture Layers"


class MyGame(arcade.Window):
    def __init__(self, width, height, title):
        super().__init__(width, height, title, resizable=True)
        self.shadertoy = Shadertoy(
            self.get_framebuffer_size(),
            """
                void mainImage( out vec4 fragColor, in vec2 fragCoord )
                {
                    // Calculate the texture coordinate of the current fragment.
                    // This interpolates from 0,0 to 1,1 from lower left to upper right
                    vec2 uv = fragCoord.xy / iResolution.xy;

                    // Write the fragment color
                    vec4 layer_0 = texture(iChannel0, uv);
                    vec4 layer_1 = texture(iChannel1, uv);
                    float value = (sin((iTime + uv.x) * 3.0) + 1.0) / 2.0;
                    fragColor = mix(layer_0, layer_1, value);
                }
            """,
        )
        # Add two OpenGL textures to different channels
        self.shadertoy.channel_0 = self.ctx.load_texture(
            ":resources:images/backgrounds/abstract_1.jpg"
        )
        self.shadertoy.channel_1 = self.ctx.load_texture(
            ":resources:images/backgrounds/abstract_2.jpg"
        )

    def on_draw(self):
        self.clear()
        self.shadertoy.render()

    def on_update(self, delta_time: float):
        self.shadertoy.time += delta_time

    def on_resize(self, width: int, height: int):
        super().on_resize(width, height)
        self.shadertoy.resize(self.get_framebuffer_size())


if __name__ == "__main__":
    MyGame(SCREEN_WIDTH, SCREEN_HEIGHT, SCREEN_TITLE)
    arcade.run()



"""
This example procedurally develops a random cave based on cellular automata.

For more information, see:
https://gamedevelopment.tutsplus.com/tutorials/generate-random-cave-levels-using-cellular-automata--gamedev-9664

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.procedural_caves_cellular
"""

import random
import arcade
import timeit

# Sprite scaling. Make this larger, like 0.5 to zoom in and add
# 'mystery' to what you can see. Make it smaller, like 0.1 to see
# more of the map.
SPRITE_SCALING = 0.25
SPRITE_SIZE = 128 * SPRITE_SCALING

# How big the grid is
GRID_WIDTH = 450
GRID_HEIGHT = 400

# Parameters for cellular automata
CHANCE_TO_START_ALIVE = 0.4
DEATH_LIMIT = 3
BIRTH_LIMIT = 4
NUMBER_OF_STEPS = 4

# How fast the player moves
MOVEMENT_SPEED = 5

# How close the player can get to the edge before we scroll.
VIEWPORT_MARGIN = 300

# How big the window is
WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Procedural Caves Cellular Automata Example"

# How fast the camera pans to the player. 1.0 is instant.
CAMERA_SPEED = 0.1


def create_grid(width, height):
    """ Create a two-dimensional grid of specified size. """
    return [[0 for _x in range(width)] for _y in range(height)]


def initialize_grid(grid):
    """ Randomly set grid locations to on/off based on chance. """
    height = len(grid)
    width = len(grid[0])
    for row in range(height):
        for column in range(width):
            if random.random() <= CHANCE_TO_START_ALIVE:
                grid[row][column] = 1


def count_alive_neighbors(grid, x, y):
    """ Count neighbors that are alive. """
    height = len(grid)
    width = len(grid[0])
    alive_count = 0
    for i in range(-1, 2):
        for j in range(-1, 2):
            neighbor_x = x + i
            neighbor_y = y + j
            if i == 0 and j == 0:
                continue
            elif (
                neighbor_x < 0 or
                neighbor_y < 0 or
                neighbor_y >= height or
                neighbor_x >= width
            ):
                # Edges are considered alive. Makes map more likely to appear naturally closed.
                alive_count += 1
            elif grid[neighbor_y][neighbor_x] == 1:
                alive_count += 1
    return alive_count


def do_simulation_step(old_grid):
    """ Run a step of the cellular automaton. """
    height = len(old_grid)
    width = len(old_grid[0])
    new_grid = create_grid(width, height)
    for x in range(width):
        for y in range(height):
            alive_neighbors = count_alive_neighbors(old_grid, x, y)
            if old_grid[y][x] == 1:
                if alive_neighbors < DEATH_LIMIT:
                    new_grid[y][x] = 0
                else:
                    new_grid[y][x] = 1
            else:
                if alive_neighbors > BIRTH_LIMIT:
                    new_grid[y][x] = 1
                else:
                    new_grid[y][x] = 0
    return new_grid


class InstructionView(arcade.View):
    """ View to show instructions """

    def __init__(self):
        super().__init__()
        self.frame_count = 0

    def on_show_view(self):
        """ This is run once when we switch to this view """
        self.window.background_color = arcade.csscolor.DARK_SLATE_BLUE

        # Reset the viewport, necessary if we have a scrolling game and we need
        # to reset the viewport back to the start so we can see what we draw.
        self.window.default_camera.use()

    def on_draw(self):
        """ Draw this view """
        self.clear()
        arcade.draw_text("Loading...", self.window.width // 2, self.window.height // 2,
                         arcade.color.BLACK, font_size=50, anchor_x="center")

    def on_update(self, dt):
        if self.frame_count == 0:
            self.frame_count += 1
            return

        """ If the user presses the mouse button, start the game. """
        game_view = GameView()
        game_view.setup()
        self.window.show_view(game_view)


class GameView(arcade.View):
    """
    Main application class.
    """

    def __init__(self):
        super().__init__()

        self.grid = None
        self.wall_list = None
        self.player_list = None
        self.player_sprite = None
        self.draw_time = 0
        self.processing_time = 0
        self.physics_engine = None

        # Track the current state of what key is pressed
        self.left_pressed = False
        self.right_pressed = False
        self.up_pressed = False
        self.down_pressed = False

        self.sprite_count_text = None
        self.draw_time_text = None
        self.processing_time_text = None

        # Create the cameras. One for the GUI, one for the sprites.
        # We scroll the 'sprite world' but not the GUI.
        self.camera_sprites = arcade.Camera2D()
        self.camera_gui = arcade.Camera2D()

        self.window.background_color = arcade.color.BLACK

    def setup(self):
        self.wall_list = arcade.SpriteList(use_spatial_hash=True)
        self.player_list = arcade.SpriteList()

        # Create cave system using a 2D grid
        self.grid = create_grid(GRID_WIDTH, GRID_HEIGHT)
        initialize_grid(self.grid)
        for step in range(NUMBER_OF_STEPS):
            self.grid = do_simulation_step(self.grid)

        texture = arcade.load_texture(":resources:images/tiles/grassCenter.png")
        # Create sprites based on 2D grid
        # Each grid location is a sprite.
        for row in range(GRID_HEIGHT):
            for column in range(GRID_WIDTH):
                if self.grid[row][column] == 1:
                    wall = arcade.BasicSprite(texture, scale=SPRITE_SCALING)
                    wall.center_x = column * SPRITE_SIZE + SPRITE_SIZE / 2
                    wall.center_y = row * SPRITE_SIZE + SPRITE_SIZE / 2
                    self.wall_list.append(wall)

        # Set up the player
        self.player_sprite = arcade.Sprite(
            ":resources:images/animated_characters/female_person/femalePerson_idle.png",
            scale=SPRITE_SCALING)
        self.player_list.append(self.player_sprite)

        # Randomly place the player. If we are in a wall, repeat until we aren't.
        placed = False
        while not placed:

            # Randomly position
            max_x = int(GRID_WIDTH * SPRITE_SIZE)
            max_y = int(GRID_HEIGHT * SPRITE_SIZE)
            self.player_sprite.center_x = random.randrange(max_x)
            self.player_sprite.center_y = random.randrange(max_y)

            # Are we in a wall?
            walls_hit = arcade.check_for_collision_with_list(self.player_sprite, self.wall_list)
            if len(walls_hit) == 0:
                # Not in a wall! Success!
                placed = True

        # Draw info on the screen
        sprite_count = len(self.wall_list)
        output = f"Sprite Count: {sprite_count:,}"
        self.sprite_count_text = arcade.Text(output,
                                             20,
                                             self.window.height - 20,
                                             arcade.color.WHITE, 16)

        output = "Drawing time:"
        self.draw_time_text = arcade.Text(output,
                                          20,
                                          self.window.height - 40,
                                          arcade.color.WHITE, 16)

        output = "Processing time:"
        self.processing_time_text = arcade.Text(output,
                                                20,
                                                self.window.height - 60,
                                                arcade.color.WHITE, 16)

        self.physics_engine = arcade.PhysicsEngineSimple(self.player_sprite,
                                                         self.wall_list)

        self.scroll_to_player(1.0)

    def on_draw(self):
        """ Render the screen. """

        # Start timing how long this takes
        draw_start_time = timeit.default_timer()

        # This command should happen before we start drawing. It will clear
        # the screen to the background color, and erase what we drew last frame.
        self.clear()

        # Select the camera we'll use to draw all our sprites
        self.camera_sprites.use()

        # Draw the sprites
        self.wall_list.draw(pixelated=True)
        self.player_list.draw()

        # Select the (unscrolled) camera for our GUI
        self.camera_gui.use()

        self.sprite_count_text.draw()
        output = f"Drawing time: {self.draw_time:.3f}"
        self.draw_time_text.text = output
        self.draw_time_text.draw()

        output = f"Processing time: {self.processing_time:.3f}"
        self.processing_time_text.text = output
        self.processing_time_text.draw()

        self.draw_time = timeit.default_timer() - draw_start_time

    def update_player_speed(self):

        # Calculate speed based on the keys pressed
        self.player_sprite.change_x = 0
        self.player_sprite.change_y = 0

        if self.up_pressed and not self.down_pressed:
            self.player_sprite.change_y = MOVEMENT_SPEED
        elif self.down_pressed and not self.up_pressed:
            self.player_sprite.change_y = -MOVEMENT_SPEED
        if self.left_pressed and not self.right_pressed:
            self.player_sprite.change_x = -MOVEMENT_SPEED
        elif self.right_pressed and not self.left_pressed:
            self.player_sprite.change_x = MOVEMENT_SPEED

    def on_key_press(self, key, modifiers):
        """Called whenever a key is pressed. """

        if key == arcade.key.UP:
            self.up_pressed = True
        elif key == arcade.key.DOWN:
            self.down_pressed = True
        elif key == arcade.key.LEFT:
            self.left_pressed = True
        elif key == arcade.key.RIGHT:
            self.right_pressed = True

    def on_key_release(self, key, modifiers):
        """Called when the user releases a key. """

        if key == arcade.key.UP:
            self.up_pressed = False
        elif key == arcade.key.DOWN:
            self.down_pressed = False
        elif key == arcade.key.LEFT:
            self.left_pressed = False
        elif key == arcade.key.RIGHT:
            self.right_pressed = False

    def scroll_to_player(self, camera_speed):
        """
        Scroll the window to the player.

        if camera_speed is 1, the camera will immediately move to the desired position.
        Anything between 0 and 1 will have the camera move to the location with a smoother
        pan.
        """

        position = (self.player_sprite.center_x, self.player_sprite.center_y)
        self.camera_sprites.position = arcade.math.lerp_2d(
            self.camera_sprites.position,
            position,
            camera_speed,
        )

    def on_resize(self, width: int, height: int):
        """
        Resize window
        Handle the user grabbing the edge and resizing the window.
        """
        super().on_resize(width, height)
        self.camera_sprites.match_window()
        self.camera_gui.match_window()

    def on_update(self, delta_time):
        """ Movement and game logic """

        start_time = timeit.default_timer()

        # Call update on all sprites (The sprites don't do much in this
        # example though.)
        self.update_player_speed()
        self.physics_engine.update()

        # Scroll the screen to the player
        self.scroll_to_player(camera_speed=CAMERA_SPEED)

        # Save the time it took to do this.
        self.processing_time = timeit.default_timer() - start_time


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create the InstructionView
    game = InstructionView()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



from __future__ import annotations

import weakref
from typing import TYPE_CHECKING

from pyglet import gl

if TYPE_CHECKING:
    from arcade.gl import Context


class Query:
    """
    A query object to perform low level measurements of OpenGL rendering calls.

    The best way to create a program instance is through :py:meth:`arcade.gl.Context.query`

    Example usage::

        query = ctx.query()
        with query:
            geometry.render(..)

        print('samples_passed:', query.samples_passed)
        print('time_elapsed:', query.time_elapsed)
        print('primitives_generated:', query.primitives_generated)

    Args:
        ctx:
            The context this query object belongs to
        samples:
            Enable counting written samples
        time:
            Enable measuring time elapsed
        primitives:
            Enable counting primitives
    """

    __slots__ = (
        "_ctx",
        "_glo_samples_passed",
        "_glo_time_elapsed",
        "_glo_primitives_generated",
        "__weakref__",
        "_samples_enabled",
        "_time_enabled",
        "_primitives_enabled",
        "_samples",
        "_time",
        "_primitives",
    )

    def __init__(self, ctx: Context, samples=True, time=True, primitives=True):
        # TODO: Support querying a subset of these queries (faster)
        # TODO: Evaluate of this query should be included
        # gl.GL_TRANSFORM_FEEDBACK_PRIMITIVES_WRITTEN
        # gl.GL_ANY_SAMPLES_PASSED
        self._ctx = ctx

        self._samples_enabled = samples
        self._time_enabled = time
        self._primitives_enabled = primitives

        self._samples = 0
        self._time = 0
        self._primitives = 0

        glos = []

        self._glo_samples_passed = glo_samples_passed = gl.GLuint()
        if self._samples_enabled:
            gl.glGenQueries(1, self._glo_samples_passed)
            glos.append(glo_samples_passed)

        self._glo_time_elapsed = glo_time_elapsed = gl.GLuint()
        if self._time_enabled:
            gl.glGenQueries(1, self._glo_time_elapsed)
            glos.append(glo_time_elapsed)

        self._glo_primitives_generated = glo_primitives_generated = gl.GLuint()
        if self._primitives_enabled:
            gl.glGenQueries(1, self._glo_primitives_generated)
            glos.append(glo_primitives_generated)

        self.ctx.stats.incr("query")

        if self._ctx.gc_mode == "auto":
            weakref.finalize(self, Query.delete_glo, self._ctx, glos)

    def __del__(self):
        if self._ctx.gc_mode == "context_gc":
            self._ctx.objects.append(self)

    @property
    def ctx(self) -> Context:
        """The context this query object belongs to"""
        return self._ctx

    @property
    def samples_passed(self) -> int:
        """
        How many samples was written. These are per component (RGBA)

        If one RGBA pixel is written, this will be 4.
        """
        return self._samples

    @property
    def time_elapsed(self) -> int:
        """The time elapsed in nanoseconds"""
        return self._time

    @property
    def primitives_generated(self) -> int:
        """
        How many primitives a vertex or geometry shader processed.

        When using a geometry shader this only counts
        the primitives actually emitted.
        """
        return self._primitives

    def __enter__(self):
        if self._ctx.gl_api == "gl":
            if self._samples_enabled:
                gl.glBeginQuery(gl.GL_SAMPLES_PASSED, self._glo_samples_passed)
            if self._time_enabled:
                gl.glBeginQuery(gl.GL_TIME_ELAPSED, self._glo_time_elapsed)
        if self._primitives_enabled:
            gl.glBeginQuery(gl.GL_PRIMITIVES_GENERATED, self._glo_primitives_generated)

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._ctx.gl_api == "gl":
            if self._samples_enabled:
                gl.glEndQuery(gl.GL_SAMPLES_PASSED)
                value = gl.GLint()
                gl.glGetQueryObjectiv(self._glo_samples_passed, gl.GL_QUERY_RESULT, value)
                self._samples = value.value

            if self._time_enabled:
                gl.glEndQuery(gl.GL_TIME_ELAPSED)
                value = gl.GLint()
                gl.glGetQueryObjectiv(self._glo_time_elapsed, gl.GL_QUERY_RESULT, value)
                self._time = value.value

        if self._primitives_enabled:
            gl.glEndQuery(gl.GL_PRIMITIVES_GENERATED)
            value = gl.GLint()
            gl.glGetQueryObjectiv(self._glo_primitives_generated, gl.GL_QUERY_RESULT, value)
            self._primitives = value.value

    def delete(self):
        """
        Destroy the underlying OpenGL resource.

        Don't use this unless you know exactly what you are doing.
        """
        Query.delete_glo(
            self._ctx,
            [
                self._glo_samples_passed,
                self._glo_time_elapsed,
                self._glo_primitives_generated,
            ],
        )

    @staticmethod
    def delete_glo(ctx, glos) -> None:
        """
        Delete this query object.

        This is automatically called when the object is garbage collected.
        """
        if gl.current_context is None:
            return

        for glo in glos:
            gl.glDeleteQueries(1, glo)

        ctx.stats.decr("query")



from __future__ import annotations

from arcade.gui.events import UIEvent, UITextInputEvent
from arcade.gui.surface import Surface
from arcade.gui.widgets.text import UIInputText


class UIPasswordInput(UIInputText):
    """A password input field. The text is hidden with asterisks.

    Hint: It is recommended to set a background color to prevent full render cycles
    when the caret blinks.

    """

    def on_event(self, event: UIEvent) -> bool | None:
        """Remove new lines from the input, which are not allowed in passwords."""
        if isinstance(event, UITextInputEvent):
            event.text = event.text.replace("\n", "").replace("\r", "")
        return super().on_event(event)

    def do_render(self, surface: Surface):
        """Override to render the text as asterisks."""
        self.layout.begin_update()
        position = self.caret.position
        text = self.text
        self.text = "*" * len(self.text)
        super().do_render(surface)
        self.text = text
        self.caret.position = position
        self.layout.end_update()
